import asyncio
from contextvars import ContextVar
import itertools

from .types import TaskGroupError

__all__ = (
    'TaskGroup',
    'current_taskgroup',
)

current_taskgroup: ContextVar['TaskGroup'] = ContextVar('current_taskgroup')


class TaskGroup(asyncio.TaskGroup):

    def __init__(self, *, name=None):
        super().__init__()
        if name is None:
            self._name = f"tg-{_name_counter()}"
        else:
            self._name = str(name)

    def get_name(self):
        return self._name

    async def __aenter__(self):
        self._current_taskgroup_token = current_taskgroup.set(self)
        return await super().__aenter__()

    async def __aexit__(self, et, exc, tb):
        try:
            return await super().__aexit__(et, exc, tb)
        except BaseExceptionGroup as eg:
            # Just wrap the exception group as TaskGroupError for backward
            # compatibility.  In Python 3.11 or higher, TaskGroupError
            # also inherits BaseExceptionGroup, so the standard except*
            # clauses and ExceptionGroup methods will work as expected.
            # New codes should migrate to directly using asyncio.TaskGroup.
            raise TaskGroupError(eg.message, eg.exceptions) from None
        finally:
            current_taskgroup.reset(self._current_taskgroup_token)


_name_counter = itertools.count(1).__next__



"""
This module implements a simple :func:`os.fork()`-like interface,
but in an asynchronous way with full support for PID file descriptors
on Python 3.9 or higher and the Linux kernel 5.4 or higher.

It internally synchronizes the beginning and readiness status of child processes
so that the users may assume that the child process is completely interruptible after
:func:`afork()` returns.
"""

import asyncio
import errno
import logging
import os
import signal
from abc import ABCMeta, abstractmethod
from typing import Callable, Tuple

from .compat import get_running_loop

__all__ = (
    'AbstractChildProcess',
    'PosixChildProcess',
    'PidfdChildProcess',
    'afork',
)

logger = logging.getLogger(__name__)

_has_pidfd = False
if hasattr(os, 'pidfd_open'):
    # signal.pidfd_send_signal() is available in Linux kernel 5.1+
    # and os.pidfd_open() is available in Linux kernel 5.3+.
    # So let's check with os.pidfd_open() which requires higher version.
    try:
        os.pidfd_open(0, 0)  # type: ignore
    except OSError as e:
        if e.errno in (errno.EBADF, errno.EINVAL):
            _has_pidfd = True
        # if the kernel does not support this,
        # it will say errno.ENOSYS or errno.EPERM


class AbstractChildProcess(metaclass=ABCMeta):
    """
    The abstract interface to control and monitor a forked child process.
    """

    @abstractmethod
    def send_signal(self, signum: int) -> None:
        """
        Send a UNIX signal to the child process.
        If the child process is already terminated, it will log
        a warning message and return.
        """
        raise NotImplementedError

    @abstractmethod
    async def wait(self) -> int:
        """
        Wait until the child process terminates or reclaim the child process' exit
        code if already terminated.
        If there are other coroutines that has waited the same process, it may
        return 255 and log a warning message.
        """
        raise NotImplementedError


class PosixChildProcess(AbstractChildProcess):
    """
    A POSIX-compatible version of :class:`AbstractChildProcess`.
    """

    def __init__(self, pid: int) -> None:
        self._pid = pid
        self._terminated = False

    def send_signal(self, signum: int) -> None:
        if self._terminated:
            if signum != signal.SIGKILL:
                logger.warning(
                    "PosixChildProcess(%d).send_signal(%d): "
                    "The process has already terminated.",
                    self._pid,
                    signum,
                )
            return
        if signum == signal.SIGKILL:
            logger.warning("Force-killed hanging child: %d", self._pid)
        os.kill(self._pid, signum)

    async def wait(self) -> int:
        loop = get_running_loop()
        try:
            _, status = await loop.run_in_executor(None, os.waitpid, self._pid, 0)
        except ChildProcessError:
            # The child process is already reaped
            # (may happen if waitpid() is called elsewhere).
            self._returncode = 255
            logger.warning(
                "child process pid %d exit status already read: "
                "it will report returncode 255",
                self._pid)
        else:
            if os.WIFSIGNALED(status):
                self._returncode = -os.WTERMSIG(status)
            elif os.WIFEXITED(status):
                self._returncode = os.WEXITSTATUS(status)
            else:
                self._returncode = status
        finally:
            self._terminated = True
        return self._returncode


class PidfdChildProcess(AbstractChildProcess):
    """
    A PID file descriptor-based version of :class:`AbstractChildProcess`.
    """

    def __init__(self, pid: int, pidfd: int) -> None:
        self._pid = pid
        self._pidfd = pidfd
        self._returncode = None
        self._wait_event = asyncio.Event()
        self._terminated = False
        loop = get_running_loop()
        loop.add_reader(self._pidfd, self._do_wait)

    def send_signal(self, signum: int) -> None:
        if self._terminated:
            if signum != signal.SIGKILL:
                logger.warning(
                    "PidfdChildProcess(%d, %d).send_signal(%d): "
                    "The process has already terminated.",
                    self._pid,
                    self._pidfd,
                    signum,
                )
            return
        if signum == signal.SIGKILL:
            logger.warning("Force-killed hanging child: %d", self._pid)
        signal.pidfd_send_signal(self._pidfd, signum)  # type: ignore

    def _do_wait(self):
        loop = get_running_loop()
        try:
            # The flag is WEXITED | __WALL from linux/wait.h
            # (__WCLONE value is out of range of int...)
            status_info = os.waitid(
                os.P_PIDFD,
                self._pidfd,
                os.WEXITED | 0x40000000,
            )
        except ChildProcessError:
            # The child process is already reaped
            # (may happen if waitpid() is called elsewhere).
            self._returncode = 255
            logger.warning(
                "child process %d exit status already read: "
                "it will report returncode 255",
                self._pid)
        else:
            if status_info.si_code == os.CLD_KILLED:
                self._returncode = -status_info.si_status  # signal number
            elif status_info.si_code == os.CLD_EXITED:
                self._returncode = status_info.si_status
            elif status_info.si_code == os.CLD_DUMPED:
                self._returncode = -status_info.si_status  # signal number
            else:
                logger.warning(
                    "unexpected si_code %d and si_status %d for child process %d",
                    status_info.si_code, status_info.si_status, self._pid)
                self._returncode = 255
        finally:
            loop.remove_reader(self._pidfd)
            os.close(self._pidfd)
            self._terminated = True
            self._wait_event.set()

    async def wait(self) -> int:
        await self._wait_event.wait()
        assert self._returncode is not None
        return self._returncode


def _child_main(init_func, init_pipe, child_func: Callable[[], int]) -> int:
    if init_func is not None:
        init_func()
    # notify the parent that the child is ready to execute the requested function.
    os.write(init_pipe, b"\0")
    os.close(init_pipe)
    return child_func()


async def _fork_posix(child_func: Callable[[], int]) -> int:
    loop = get_running_loop()
    init_pipe = os.pipe()

    pid = os.fork()
    if pid == 0:
        os.close(init_pipe[0])
        ret = 0
        try:
            ret = _child_main(None, init_pipe[1], child_func)
        except KeyboardInterrupt:
            ret = -signal.SIGINT
        finally:
            os._exit(ret)
        return ret
    os.close(init_pipe[1])

    # Wait for the child's readiness notification
    init_event = asyncio.Event()
    loop.add_reader(init_pipe[0], init_event.set)
    await init_event.wait()
    loop.remove_reader(init_pipe[0])
    os.read(init_pipe[0], 1)
    os.close(init_pipe[0])
    return pid


async def _clone_pidfd(child_func: Callable[[], int]) -> Tuple[int, int]:
    loop = get_running_loop()
    init_pipe = os.pipe()

    pid = os.fork()
    if pid == 0:
        os.close(init_pipe[0])
        ret = 0
        try:
            ret = _child_main(None, init_pipe[1], child_func)
        except KeyboardInterrupt:
            ret = -signal.SIGINT
        finally:
            os._exit(ret)
        return ret
    os.close(init_pipe[1])

    # Get the pidfd.
    fd = os.pidfd_open(pid, 0)  # type: ignore

    # Wait for the child's readiness notification
    init_event = asyncio.Event()
    loop.add_reader(init_pipe[0], init_event.set)
    await init_event.wait()
    loop.remove_reader(init_pipe[0])
    os.read(init_pipe[0], 1)
    os.close(init_pipe[0])
    return pid, fd


async def afork(child_func: Callable[[], int]) -> AbstractChildProcess:
    """
    Fork the current process and execute the given function in the child.
    The return value of the function will become the exit code of the child
    process.

    Args:
        child_func: A function that represents the main function of the child and
                    returns an integer as its exit code.
                    Note that the function must set up a new event loop if it
                    wants to run asyncio codes.
    """
    if _has_pidfd:
        pid, pidfd = await _clone_pidfd(child_func)
        return PidfdChildProcess(pid, pidfd)
    else:
        pid = await _fork_posix(child_func)
        return PosixChildProcess(pid)



import asyncio
import itertools
import logging
import sys
import traceback
import weakref
from contextvars import ContextVar, Token
from types import TracebackType
from typing import (
    Any,
    Awaitable,
    Callable,
    Coroutine,
    Optional,
    Sequence,
    Type,
    Union,
)

from .. import compat
from .types import AsyncExceptionHandler

__all__ = (
    'PersistentTaskGroup',
    'current_ptaskgroup',
)

current_ptaskgroup: ContextVar['PersistentTaskGroup'] = \
    ContextVar('current_ptaskgroup')

_ptaskgroup_idx = itertools.count()
_log = logging.getLogger(__name__)
_all_ptaskgroups: weakref.WeakSet['PersistentTaskGroup'] = weakref.WeakSet()


async def _default_exc_handler(exc_type, exc_obj, exc_tb) -> None:
    traceback.print_exc()


class PersistentTaskGroup:
    _base_error: Optional[BaseException]
    _exc_handler: AsyncExceptionHandler
    _tasks: "weakref.WeakSet[asyncio.Task]"
    _on_completed_fut: Optional[asyncio.Future]
    _current_taskgroup_token: Optional[Token["PersistentTaskGroup"]]

    @classmethod
    def all_ptaskgroups(cls) -> Sequence['PersistentTaskGroup']:
        return list(_all_ptaskgroups)

    def __init__(
        self,
        *,
        name: Optional[str] = None,
        exception_handler: Optional[AsyncExceptionHandler] = None,
    ) -> None:
        self._entered = False
        self._exiting = False
        self._aborting = False
        self._base_error = None
        self._name = name or f"{next(_ptaskgroup_idx)}"
        self._parent_cancel_requested = False
        self._unfinished_tasks = 0
        self._on_completed_fut = None
        self._parent_task = compat.current_task()
        self._tasks = weakref.WeakSet()
        self._current_taskgroup_token = None
        _all_ptaskgroups.add(self)
        if exception_handler is None:
            self._exc_handler = _default_exc_handler
        else:
            self._exc_handler = exception_handler

    # TODO: task statistics and enumeration (for aiomonitor)
    # TODO: two-phase shutdown

    def get_name(self) -> str:
        return self._name

    def create_task(
        self,
        coro: Coroutine[Any, Any, Any],
        *,
        name: Optional[str] = None,
    ) -> Awaitable[Any]:
        if not self._entered:
            # When used as object attribute, auto-enter.
            self._entered = True
        if self._exiting and self._unfinished_tasks == 0:
            raise RuntimeError(f"{self!r} has already finished")
        return self._create_task_with_name(coro, name=name, cb=self._on_task_done)

    def _create_task_with_name(
        self,
        coro: Coroutine[Any, Any, Any],
        *,
        name: Optional[str] = None,
        cb: Callable[[asyncio.Task], Any],
    ) -> Awaitable[Any]:
        loop = compat.get_running_loop()
        result_future = loop.create_future()
        child_task = loop.create_task(
            self._task_wrapper(coro, weakref.ref(result_future)),
            name=name,
        )
        _log.debug("%r is spawned in %r.", child_task, self)
        self._unfinished_tasks += 1
        child_task.add_done_callback(cb)
        self._tasks.add(child_task)
        return result_future

    def _is_base_error(self, exc: BaseException) -> bool:
        assert isinstance(exc, BaseException)
        return isinstance(exc, (SystemExit, KeyboardInterrupt))

    async def _wait_completion(self) -> Optional[BaseException]:
        loop = compat.get_running_loop()
        propagate_cancellation_error = None
        while self._unfinished_tasks:
            if self._on_completed_fut is None:
                self._on_completed_fut = loop.create_future()
            try:
                await self._on_completed_fut
            except asyncio.CancelledError as ex:
                if not self._aborting:
                    propagate_cancellation_error = ex
                    self._trigger_shutdown()
            self._on_completed_fut = None

        assert self._unfinished_tasks == 0
        self._on_completed_fut = None
        _all_ptaskgroups.discard(self)
        return propagate_cancellation_error

    def _trigger_shutdown(self) -> None:
        self._aborting = True
        for t in self._tasks:
            if not t.done():
                t.cancel()

    async def shutdown(self) -> None:
        self._trigger_shutdown()
        await self._wait_completion()

    async def _task_wrapper(
        self,
        coro: Coroutine,
        result_future: weakref.ref[asyncio.Future],
    ) -> Any:
        loop = compat.get_running_loop()
        task = compat.current_task()
        fut = result_future()
        try:
            ret = await coro
            if fut is not None:
                fut.set_result(ret)
            return ret
        except asyncio.CancelledError:
            if fut is not None:
                fut.cancel()
            raise
        except Exception as e:
            # Swallow unhandled exceptions by our own and
            # prevent abortion of the task group bu them.
            # Wrapping corotuines directly has advantage for
            # exception handlers to access full traceback
            # and there is no need to implement separate
            # mechanism to wait for exception handler tasks.
            try:
                if fut is not None:
                    fut.set_exception(e)
                await self._exc_handler(*sys.exc_info())
            except Exception as exc:
                # If there are exceptions inside the exception handler
                # we report it as soon as possible using the event loop's
                # exception handler, instead of postponing
                # to the timing when PersistentTaskGroup terminates.
                loop.call_exception_handler({
                    'message': f"Got an unhandled exception "
                               f"in the exception handler of Task {task!r}",
                    'exception': exc,
                    'task': task,
                })
        finally:
            del fut

    def _on_task_done(self, task: asyncio.Task) -> None:
        self._unfinished_tasks -= 1
        assert self._unfinished_tasks >= 0
        assert self._parent_task is not None

        if self._on_completed_fut is not None and not self._unfinished_tasks:
            if not self._on_completed_fut.done():
                self._on_completed_fut.set_result(True)

        if task.cancelled():
            _log.debug("%r in %r has been cancelled.", task, self)
            return

        exc = task.exception()
        if exc is None:
            return

        # Now the exception is BaseException.
        if self._base_error is None:
            self._base_error = exc

        self._trigger_shutdown()
        if not self._parent_task.cancelling():
            self._parent_cancel_requested = True

    async def __aenter__(self) -> "PersistentTaskGroup":
        self._parent_task = compat.current_task()
        self._current_taskgroup_token = current_ptaskgroup.set(self)
        self._entered = True
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> Optional[bool]:
        assert self._parent_task is not None
        self._exiting = True
        propagate_cancellation_error: Optional[
            Union[Type[BaseException], BaseException]
        ] = None

        if (exc_val is not None and
                self._is_base_error(exc_val) and
                self._base_error is None):
            self._base_error = exc_val

        if (
            exc_type is asyncio.CancelledError or
            exc_type is asyncio.TimeoutError
        ):
            if self._parent_cancel_requested:
                self._parent_task.uncancel()
            else:
                propagate_cancellation_error = exc_type
        if exc_type is not None and not self._aborting:
            if (
                exc_type is asyncio.CancelledError or
                exc_type is asyncio.TimeoutError
            ):
                propagate_cancellation_error = exc_type
            self._trigger_shutdown()

        prop_ex = await self._wait_completion()
        if prop_ex is not None:
            propagate_cancellation_error = prop_ex
        if self._current_taskgroup_token:
            current_ptaskgroup.reset(self._current_taskgroup_token)
            self._current_taskgroup_token = None
        if propagate_cancellation_error is not None:
            raise propagate_cancellation_error

        return None

    def __repr__(self) -> str:
        info = ['']
        if self._name:
            info.append(f'name={self._name}')
        if self._tasks:
            info.append(f'tasks={len(self._tasks)}')
        if self._unfinished_tasks:
            info.append(f'unfinished={self._unfinished_tasks}')
        if self._aborting:
            info.append('cancelling')
        elif self._entered:
            info.append('entered')
        info_str = ' '.join(info)
        return f'<PersistentTaskGroup({info_str})>'



import asyncio
import os
import signal
import time
from unittest import mock

import pytest

from aiotools import fork as fork_mod
from aiotools.fork import afork, _has_pidfd, PidfdChildProcess


async def _do_test_fork():

    def child():
        print("hello world")
        time.sleep(0.1)
        return 99

    proc = await afork(child)
    assert proc._pid > 0
    if isinstance(proc, PidfdChildProcess):
        assert proc._pidfd > 0
    ret = await proc.wait()
    assert ret == 99


async def _do_test_fork_already_terminated():

    def child():
        time.sleep(0.1)
        return 99

    proc = await afork(child)
    assert proc._pid > 0
    if isinstance(proc, PidfdChildProcess):
        assert proc._pidfd > 0
    await asyncio.sleep(0.5)
    ret = await proc.wait()
    assert ret == 99


async def _do_test_fork_signal():

    def child():
        try:
            time.sleep(10)
        except KeyboardInterrupt:
            return 101
        return 100

    os.setpgrp()
    proc = await afork(child)
    assert proc._pid > 0
    if isinstance(proc, PidfdChildProcess):
        assert proc._pidfd > 0
    await asyncio.sleep(0.1)
    proc.send_signal(signal.SIGINT)
    ret = await proc.wait()
    # FIXME: Sometimes it returns 254
    assert ret == 101


async def _do_test_fork_segfault():

    def child():
        time.sleep(0.1)
        import ctypes
        ctypes.string_at(0)  # segfault!

    os.setpgrp()
    proc = await afork(child)
    assert proc._pid > 0
    if isinstance(proc, PidfdChildProcess):
        assert proc._pidfd > 0
    ret = await proc.wait()
    assert ret == -11  # SIGSEGV


async def _do_test_fork_many():

    def child():
        try:
            time.sleep(10)
        except KeyboardInterrupt:
            return 101
        return 100

    os.setpgrp()
    proc_list = []
    for _ in range(32):
        proc = await afork(child)
        proc_list.append(proc)
        assert proc._pid > 0
        if isinstance(proc, PidfdChildProcess):
            assert proc._pidfd > 0
    for i in range(16):
        proc_list[i].send_signal(signal.SIGINT)
    for i in range(16, 32):
        proc_list[i].send_signal(signal.SIGTERM)
    ret_list = await asyncio.gather(*[proc.wait() for proc in proc_list])
    for i in range(16):
        assert ret_list[i] == 101
    for i in range(16, 32):
        assert ret_list[i] == -15  # killed by SIGTERM


@pytest.mark.skipif(
    not _has_pidfd,
    reason="pidfd is supported in Python 3.9+ and Linux kernel 5.4+"
)
@pytest.mark.asyncio
async def test_fork():
    await _do_test_fork()


@pytest.mark.skipif(
    not _has_pidfd,
    reason="pidfd is supported in Python 3.9+ and Linux kernel 5.4+"
)
@pytest.mark.asyncio
async def test_fork_already_terminated():
    await _do_test_fork_already_terminated()


@pytest.mark.skipif(
    not _has_pidfd,
    reason="pidfd is supported in Python 3.9+ and Linux kernel 5.4+"
)
@pytest.mark.asyncio
async def test_fork_signal():
    await _do_test_fork_signal()


@pytest.mark.skipif(
    not _has_pidfd,
    reason="pidfd is supported in Python 3.9+ and Linux kernel 5.4+"
)
@pytest.mark.asyncio
async def test_fork_segfault():
    await _do_test_fork_segfault()


@pytest.mark.skipif(
    not _has_pidfd,
    reason="pidfd is supported in Python 3.9+ and Linux kernel 5.4+"
)
@pytest.mark.asyncio
async def test_fork_many():
    await _do_test_fork_many()


@pytest.mark.asyncio
async def test_fork_fallback():
    with mock.patch.object(
        fork_mod, '_has_pidfd', False,
    ):
        await _do_test_fork()


@pytest.mark.asyncio
async def test_fork_already_termination_fallback():
    with mock.patch.object(
        fork_mod, '_has_pidfd', False,
    ):
        await _do_test_fork_already_terminated()


@pytest.mark.asyncio
async def test_fork_signal_fallback():
    with mock.patch.object(
        fork_mod, '_has_pidfd', False,
    ):
        await _do_test_fork_signal()


@pytest.mark.asyncio
async def test_fork_segfault_fallback():
    with mock.patch.object(
        fork_mod, '_has_pidfd', False,
    ):
        await _do_test_fork_segfault()


@pytest.mark.asyncio
async def test_fork_many_fallback():
    with mock.patch.object(
        fork_mod, '_has_pidfd', False,
    ):
        await _do_test_fork_many()



from pprint import pprint
import asyncio
import aiotools


@aiotools.actxmgr
async def mygen(id):
    yield f'mygen id is {id}'


async def run():
    ctxgrp = aiotools.actxgroup(mygen(i) for i in range(10))
    async with ctxgrp as values:
        pprint(values)


if __name__ == '__main__':
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    try:
        loop.run_until_complete(run())
    finally:
        loop.stop()



"""
This module is a modified version of :mod:`asyncio.timeouts` from Python 3.11 stdlib,
which looks into :class:`BaseExceptionGroup` if raised from the inner block of the
timeout context manager.
"""

import enum

from types import TracebackType
from typing import final, Optional, Type

from asyncio import events
from asyncio import exceptions
from asyncio import tasks


__all__ = (
    "Timeout",
    "timeout",
    "timeout_at",
)


class _State(enum.Enum):
    CREATED = "created"
    ENTERED = "active"
    EXPIRING = "expiring"
    EXPIRED = "expired"
    EXITED = "finished"


@final
class Timeout:
    """Asynchronous context manager for cancelling overdue coroutines.

    Use `timeout()` or `timeout_at()` rather than instantiating this class directly.
    """

    def __init__(self, when: Optional[float]) -> None:
        """Schedule a timeout that will trigger at a given loop time.

        - If `when` is `None`, the timeout will never trigger.
        - If `when < loop.time()`, the timeout will trigger on the next
          iteration of the event loop.
        """
        self._state = _State.CREATED

        self._timeout_handler: Optional[events.TimerHandle] = None
        self._task: Optional[tasks.Task] = None
        self._when = when

    def when(self) -> Optional[float]:
        """Return the current deadline."""
        return self._when

    def reschedule(self, when: Optional[float]) -> None:
        """Reschedule the timeout."""
        assert self._state is not _State.CREATED
        if self._state is not _State.ENTERED:
            raise RuntimeError(
                f"Cannot change state of {self._state.value} Timeout",
            )

        self._when = when

        if self._timeout_handler is not None:
            self._timeout_handler.cancel()

        if when is None:
            self._timeout_handler = None
        else:
            loop = events.get_running_loop()
            if when <= loop.time():
                self._timeout_handler = (
                    loop.call_soon(self._on_timeout)  # type: ignore
                )
            else:
                self._timeout_handler = loop.call_at(when, self._on_timeout)

    def expired(self) -> bool:
        """Is timeout expired during execution?"""
        return self._state in (_State.EXPIRING, _State.EXPIRED)

    def __repr__(self) -> str:
        info = ['']
        if self._state is _State.ENTERED:
            when = round(self._when, 3) if self._when is not None else None
            info.append(f"when={when}")
        info_str = ' '.join(info)
        return f"<Timeout [{self._state.value}]{info_str}>"

    async def __aenter__(self) -> "Timeout":
        self._state = _State.ENTERED
        self._task = tasks.current_task()
        assert self._task is not None
        self._cancelling = self._task.cancelling()
        if self._task is None:
            raise RuntimeError("Timeout should be used inside a task")
        self.reschedule(self._when)
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> Optional[bool]:
        assert self._state in (_State.ENTERED, _State.EXPIRING)
        assert self._task is not None

        if self._timeout_handler is not None:
            self._timeout_handler.cancel()
            self._timeout_handler = None

        if self._state is _State.EXPIRING:
            self._state = _State.EXPIRED

            contains_cancellation = False
            if isinstance(exc_val, BaseExceptionGroup):
                matched, rest = exc_val.split(exceptions.CancelledError)
                # Check and strip out the cancellation errors from inside.
                contains_cancellation = (matched is not None)
                if (
                    self._task.uncancel() <= self._cancelling and
                    contains_cancellation
                ):
                    if rest is not None:
                        raise BaseExceptionGroup(
                            "timeout with collected inner exceptions",
                            [TimeoutError(), rest],
                        )
                    raise TimeoutError from exc_val
            else:
                if (
                    self._task.uncancel() <= self._cancelling and
                    exc_type is exceptions.CancelledError
                ):
                    # Since there are no new cancel requests, we're
                    # handling this.
                    raise TimeoutError from exc_val
        elif self._state is _State.ENTERED:
            self._state = _State.EXITED

        return None

    def _on_timeout(self) -> None:
        assert self._state is _State.ENTERED
        assert self._task is not None
        self._task.cancel("timeout")
        self._state = _State.EXPIRING
        # drop the reference early
        self._timeout_handler = None


def timeout(delay: Optional[float]) -> Timeout:
    """Timeout async context manager.

    Useful in cases when you want to apply timeout logic around block
    of code or in cases when asyncio.wait_for is not suitable. For example:

    >>> async with asyncio.timeout(10):  # 10 seconds timeout
    ...     await long_running_task()


    delay - value in seconds or None to disable timeout logic

    long_running_task() is interrupted by raising asyncio.CancelledError,
    the top-most affected timeout() context manager converts CancelledError
    into TimeoutError.
    """
    loop = events.get_running_loop()
    return Timeout(loop.time() + delay if delay is not None else None)


def timeout_at(when: Optional[float]) -> Timeout:
    """Schedule the timeout at absolute time.

    Like timeout() but argument gives absolute time in the same clock system
    as loop.time().

    Please note: it is not POSIX time but a time with
    undefined starting base, e.g. the time of the system power on.

    >>> async with asyncio.timeout_at(loop.time() + 10):
    ...     await long_running_task()


    when - a deadline when timeout occurs or None to disable timeout logic

    long_running_task() is interrupted by raising asyncio.CancelledError,
    the top-most affected timeout() context manager converts CancelledError
    into TimeoutError.
    """
    return Timeout(when)



import asyncio
import sys
from unittest import mock

import pytest

import aiotools


# NOTE: Until pytest-asyncio support ExceptionGroup,
#       assertion failures inside PersistentTaskGroup/TaskGroup blocks
#       may be represented as sub-task errors instead of
#       being logged explicitly by pytest.


@pytest.mark.asyncio
async def test_ptaskgroup_naming():

    async def subtask():
        pass

    async with aiotools.PersistentTaskGroup(name="XYZ") as tg:
        t = tg.create_task(subtask(), name="ABC")
        assert tg.get_name() == "XYZ"
        if hasattr(t, 'get_name'):
            assert t.get_name() == "ABC"


@pytest.mark.asyncio
async def test_ptaskgroup_all_done():

    done_count = 0

    async def subtask():
        nonlocal done_count
        await asyncio.sleep(0.1)
        done_count += 1

    vclock = aiotools.VirtualClock()
    with vclock.patch_loop():

        async with aiotools.PersistentTaskGroup() as tg:
            for idx in range(10):
                tg.create_task(subtask())
            assert tg._unfinished_tasks == 10
            # wait until all is done
            await asyncio.sleep(0.2)
            assert done_count == 10
            assert len(tg._tasks) == 0
            assert tg._unfinished_tasks == 0

        assert done_count == 10
        with pytest.raises(RuntimeError):
            tg.create_task(subtask())


@pytest.mark.asyncio
async def test_ptaskgroup_as_obj_attr():

    done_count = 0

    async def subtask():
        nonlocal done_count
        await asyncio.sleep(0.1)
        done_count += 1

    class LongLivedObject:

        def __init__(self):
            self.tg = aiotools.PersistentTaskGroup()
            assert not self.tg._entered

        async def work(self):
            self.tg.create_task(subtask())
            assert self.tg._entered

        async def aclose(self):
            await self.tg.shutdown()

    vclock = aiotools.VirtualClock()
    with vclock.patch_loop():

        obj = LongLivedObject()
        for idx in range(10):
            await obj.work()
        assert obj.tg._unfinished_tasks == 10

        # shutdown after all done
        await asyncio.sleep(0.2)
        await obj.aclose()

        assert done_count == 10
        assert len(obj.tg._tasks) == 0
        assert obj.tg._unfinished_tasks == 0

        done_count = 0
        obj = LongLivedObject()
        for idx in range(10):
            await obj.work()
        assert obj.tg._unfinished_tasks == 10

        # shutdown immediately
        await obj.aclose()

        assert done_count == 0
        assert len(obj.tg._tasks) == 0
        assert obj.tg._unfinished_tasks == 0


@pytest.mark.asyncio
async def test_ptaskgroup_shutdown_from_different_task():

    done_count = 0
    exec_after_termination = False

    async def subtask(idx):
        nonlocal done_count
        await asyncio.sleep(0.1 * idx)
        done_count += 1

    vclock = aiotools.VirtualClock()
    with vclock.patch_loop():

        outer_myself = aiotools.compat.current_task()
        tg = aiotools.PersistentTaskGroup()
        assert tg._parent_task is outer_myself

        async def _main_task():
            nonlocal exec_after_termination
            myself = aiotools.compat.current_task()
            async with tg:
                # The parent task is overriden when
                # using "async with", to keep consistency with
                # the original asyncio.TaskGroup.
                assert tg._parent_task is myself

                for idx in range(10):
                    tg.create_task(subtask(idx))

            # The code below must be executed even when
            # tg is shutdown from other tasks.
            exec_after_termination = True

        async def _stop_task():
            await asyncio.sleep(0.49)
            await tg.shutdown()

        async with aiotools.TaskGroup() as outer_tg:
            outer_tg.create_task(_main_task())
            outer_tg.create_task(_stop_task())

        assert done_count == 5
        assert exec_after_termination
        assert tg._parent_task is not outer_myself
        assert tg._unfinished_tasks == 0


@pytest.mark.asyncio
async def test_ptaskgroup_cancel_after_schedule():

    done_count = 0

    vclock = aiotools.VirtualClock()
    with vclock.patch_loop():

        async def subtask():
            nonlocal done_count
            await asyncio.sleep(0.1)
            done_count += 1

        async with aiotools.PersistentTaskGroup() as tg:
            for _ in range(10):
                tg.create_task(subtask())
            await asyncio.sleep(0)

        # shutdown after exit (all done) is no-op.
        assert done_count == 10
        assert len(tg._tasks) == 0
        assert tg._unfinished_tasks == 0
        await tg.shutdown()
        assert len(tg._tasks) == 0
        assert tg._unfinished_tasks == 0


@pytest.mark.asyncio
async def test_ptaskgroup_cancel_before_schedule():

    done_count = 0

    async def subtask():
        nonlocal done_count
        await asyncio.sleep(0.1)
        done_count += 1

    vclock = aiotools.VirtualClock()
    with vclock.patch_loop():

        async with aiotools.PersistentTaskGroup() as tg:
            for _ in range(10):
                tg.create_task(subtask())
            # let's abort immediately.
            await tg.shutdown()

        assert done_count == 0
        assert len(tg._tasks) == 0


@pytest.mark.skipif(
    sys.version_info < (3, 7, 0),
    reason='Requires Python 3.7 or higher',
    # In Python 3.6, this test hangs indefinitely.
    # We don't fix this -- 3.6 is EoL as of December 2021.
)
@pytest.mark.asyncio
async def test_ptaskgroup_await_result():

    done_count = 0

    async def subtask():
        nonlocal done_count
        await asyncio.sleep(0.1)
        done_count += 1
        return "a"

    vclock = aiotools.VirtualClock()
    with vclock.patch_loop():

        results = []

        async with aiotools.PersistentTaskGroup() as tg:

            ret = await tg.create_task(subtask())
            results.append(ret)

            ret = await asyncio.shield(tg.create_task(subtask()))
            results.append(ret)

            a = tg.create_task(subtask())
            try:
                ret = await a
                results.append(ret)
            finally:
                del a

            a = asyncio.shield(tg.create_task(subtask()))
            try:
                ret = await a
                results.append(ret)
            finally:
                del a

        # task callbacks has an extra ref, and they need to finish before
        # we task objetcs get garbage-collected.
        await asyncio.sleep(0)
        assert results == ["a", "a", "a", "a"]
        assert done_count == 4
        assert tg._unfinished_tasks == 0
        assert len(tg._tasks) == 0


@pytest.mark.skipif(
    sys.version_info < (3, 7, 0),
    reason='Requires Python 3.7 or higher',
    # In Python 3.6, this test hangs indefinitely.
    # We don't fix this -- 3.6 is EoL as of December 2021.
)
@pytest.mark.asyncio
async def test_ptaskgroup_await_exception():

    done_count = 0
    error_count = 0

    async def subtask():
        nonlocal done_count
        await asyncio.sleep(0.1)
        1 / 0
        done_count += 1

    async def handler(exc_type, exc_obj, exc_tb):
        nonlocal error_count
        assert issubclass(exc_type, ZeroDivisionError)
        error_count += 1

    vclock = aiotools.VirtualClock()
    with vclock.patch_loop():

        async with aiotools.PersistentTaskGroup(exception_handler=handler) as tg:

            with pytest.raises(ZeroDivisionError):
                await tg.create_task(subtask())

            with pytest.raises(ZeroDivisionError):
                await asyncio.shield(tg.create_task(subtask()))

            with pytest.raises(ZeroDivisionError):
                a = tg.create_task(subtask())
                try:
                    await a
                finally:
                    del a

            with pytest.raises(ZeroDivisionError):
                # WARNING: This pattern leaks the reference to the task.
                a = asyncio.shield(tg.create_task(subtask()))
                try:
                    await a
                finally:
                    del a

        # task callbacks has an extra ref, and they need to finish before
        # we task objetcs get garbage-collected.
        await asyncio.sleep(0)
        assert done_count == 0
        assert error_count == 4
        assert tg._unfinished_tasks == 0
        assert len(tg._tasks) == 0


@pytest.mark.asyncio
async def test_ptaskgroup_exc_handler_swallow():

    done_count = 0
    error_count = 0

    async def subtask():
        nonlocal done_count
        await asyncio.sleep(0.1)
        1 / 0
        done_count += 1

    async def handler(exc_type, exc_obj, exc_tb):
        nonlocal error_count
        assert issubclass(exc_type, ZeroDivisionError)
        error_count += 1

    vclock = aiotools.VirtualClock()
    with vclock.patch_loop():

        try:
            async with aiotools.PersistentTaskGroup(exception_handler=handler) as tg:
                for _ in range(10):
                    tg.create_task(subtask())
        except ExceptionGroup as eg:
            # All non-base exceptions must be swallowed by
            # our exception handler.
            assert len(eg.subgroup(ZeroDivisionError).exceptions) == 0

        assert done_count == 0
        assert error_count == 10
        assert len(tg._tasks) == 0
        assert tg._unfinished_tasks == 0


@pytest.mark.skipif(
    sys.version_info < (3, 8, 0),
    reason='Requires Python 3.8 or higher',
)
@pytest.mark.asyncio
async def test_ptaskgroup_error_in_exc_handlers():

    done_count = 0
    error_count = 0

    async def subtask():
        nonlocal done_count
        await asyncio.sleep(0.1)
        1 / 0
        done_count += 1

    async def handler(exc_type, exc_obj, exc_tb):
        nonlocal error_count
        assert issubclass(exc_type, ZeroDivisionError)
        error_count += 1
        raise ValueError("something wrong in exception handler")

    loop = aiotools.compat.get_running_loop()
    vclock = aiotools.VirtualClock()
    with vclock.patch_loop(), \
         mock.patch.object(
             loop,
             'call_exception_handler',
             mock.MagicMock(),
         ):
        # Errors in exception handlers are covered by the event loop's exception
        # handler, so that they can be reported as soon as possible when they occur.
        #
        # In asyncio.TaskGroup, they are propagated as an exception group when
        # the task group terminates, but in PersistentTaskGroup it results in delayed
        # propagation because it needs to wait until other tasks to finish and
        # may not terminate at all until the application terminates if used as object
        # attributes instead of an async context manager.

        try:
            async with aiotools.PersistentTaskGroup(exception_handler=handler) as tg:
                for _ in range(10):
                    tg.create_task(subtask())
        except ValueError:
            assert False, "should not reach here"
        except ExceptionGroup:
            assert False, "should not reach here"

        # Check if the event loop exception handler is called.
        loop.call_exception_handler.assert_called()
        calls = loop.call_exception_handler.mock_calls
        for idx in range(10):
            assert isinstance(calls[idx].args[0]['exception'], ValueError)
        loop.call_exception_handler.reset_mock()  # to clean up task refs
        del calls  # to clean up task refs
        assert done_count == 0
        assert error_count == 10
        assert len(tg._tasks) == 0
        assert tg._unfinished_tasks == 0

        done_count = 0
        error_count = 0
        try:
            async with aiotools.PersistentTaskGroup(exception_handler=handler) as tg:
                for _ in range(10):
                    tg.create_task(subtask())
        except ExceptionGroup:
            assert False, "should not reach here"

        # Check if the event loop exception handler is called.
        loop.call_exception_handler.assert_called()
        calls = loop.call_exception_handler.mock_calls
        for idx in range(10):
            assert isinstance(calls[idx].args[0]['exception'], ValueError)
        loop.call_exception_handler.reset_mock()  # to clean up task refs
        del calls  # to clean up task refs
        assert done_count == 0
        assert error_count == 10
        assert len(tg._tasks) == 0
        assert tg._unfinished_tasks == 0


@pytest.mark.asyncio
async def test_ptaskgroup_cancel_with_await():

    done_count = 0

    async def subtask():
        nonlocal done_count
        try:
            await asyncio.sleep(0.1)
            done_count += 1   # should not be executed
        except asyncio.CancelledError:
            await asyncio.sleep(0.1)
            done_count += 10  # should be executed

    vclock = aiotools.VirtualClock()
    with vclock.patch_loop():

        async with aiotools.PersistentTaskGroup() as tg:
            for _ in range(10):
                tg.create_task(subtask())
            # Shutdown just after starting child tasks.
            # Even in this case, awaits in the tasks' cancellation blocks
            # should be executed until their completion.
            await asyncio.sleep(0.01)
            await tg.shutdown()

        # ensure that awaits in all cancellation handling blocks have been executed
        assert done_count == 100
        assert len(tg._tasks) == 0
        assert tg._unfinished_tasks == 0


@pytest.mark.skipif(
    sys.version_info < (3, 7, 0),
    reason='Requires Python 3.7 or higher',
)
@pytest.mark.asyncio
async def test_ptaskgroup_current():

    names = []

    async def subtask():
        await asyncio.sleep(1)
        names.append(aiotools.current_ptaskgroup.get().get_name())

    async def job():
        names.append(aiotools.current_ptaskgroup.get().get_name())
        async with aiotools.PersistentTaskGroup(name="inner") as tg:
            for _ in range(10):
                tg.create_task(subtask())

    vclock = aiotools.VirtualClock()
    with vclock.patch_loop():

        async with aiotools.PersistentTaskGroup(name="outer") as tg:
            tg.create_task(job())
            tg.create_task(job())
            tg.create_task(job())

    assert names == ["outer"] * 3 + ["inner"] * 30


@pytest.mark.asyncio
async def test_ptaskgroup_enumeration():

    async def subtask():
        await asyncio.sleep(1)

    async def job():
        async with aiotools.PersistentTaskGroup() as tg:
            for _ in range(10):
                tg.create_task(subtask())

    vclock = aiotools.VirtualClock()
    with vclock.patch_loop():

        async with aiotools.PersistentTaskGroup() as tg:
            tg.create_task(job())
            tg.create_task(job())
            tg.create_task(job())
            await asyncio.sleep(0.1)
            all_tgs = aiotools.PersistentTaskGroup.all_ptaskgroups()

        assert len(all_tgs) == 4



import asyncio
from contextvars import Context, ContextVar, copy_context
from typing import TypeVar

import pytest

from aiotools import VirtualClock, race

T = TypeVar("T")
cancelled = ContextVar("cancelled", default=0)


async def do_job(delay: float, result: T) -> T:
    try:
        await asyncio.sleep(delay)
        return result
    except asyncio.CancelledError:
        await asyncio.sleep(0.1)
        cancelled.set(cancelled.get() + 1)
        raise


async def fail_job(delay: float) -> None:
    try:
        await asyncio.sleep(delay)
        1 / 0
    except asyncio.CancelledError:
        await asyncio.sleep(0.1)
        cancelled.set(cancelled.get() + 1)
        raise


@pytest.mark.asyncio
async def test_race_partial_failure():
    context = copy_context()
    with VirtualClock().patch_loop():

        async def _inner1() -> None:
            result, errors = await race([
                do_job(0.1, 1),  # returns here
                fail_job(0.3),   # cancelled
                do_job(0.5, 5),  # cancelled
            ], context=context)
            assert result == 1
            assert len(errors) == 0  # should be empty
            assert cancelled.get() == 2

        # Fix the context used by all sub-tasks
        await asyncio.create_task(_inner1(), context=context)

        async def _inner2() -> None:
            with pytest.raises(ZeroDivisionError):
                await race([
                    fail_job(0.3),   # raises here
                    do_job(0.4, 1),  # cancelled
                    do_job(0.5, 5),  # cancelled
                ], context=context)
                assert cancelled.get() == 2

        # Fix the context used by all sub-tasks
        await asyncio.create_task(_inner2(), context=context)


@pytest.mark.asyncio
async def test_race_continue_on_error():
    context = copy_context()
    with VirtualClock().patch_loop():

        async def _inner() -> None:
            result, errors = await race([
                fail_job(0.3),   # collected
                fail_job(0.35),  # collected
                do_job(0.4, 4),  # returns here
                fail_job(0.45),  # cancelled
                do_job(0.5, 5),  # cancelled
            ], continue_on_error=True, context=context)
            assert result == 4
            assert len(errors) == 2
            assert isinstance(errors[0], ZeroDivisionError)
            assert isinstance(errors[1], ZeroDivisionError)
            assert cancelled.get() == 2

        await asyncio.create_task(_inner(), context=context)


@pytest.mark.asyncio
async def test_race_empty_coro_list():
    with pytest.raises(ValueError):
        await race([])


@pytest.mark.asyncio
async def test_race_all_failure():
    context = Context()
    with VirtualClock().patch_loop():
        async def _inner() -> None:
            with pytest.raises(ZeroDivisionError):
                await race([
                    fail_job(0.1),  # raises here
                    fail_job(0.3),  # cancelled
                    fail_job(0.5),  # cancelled
                ], context=context)
            assert cancelled.get() == 2

        await asyncio.create_task(_inner(), context=context)


@pytest.mark.asyncio
async def test_race_all_failure_with_continue_on_error():
    context = Context()
    with VirtualClock().patch_loop():
        async def _inner() -> None:
            try:
                await race([
                    fail_job(0.1),  # collected
                    fail_job(0.3),  # collected
                    fail_job(0.5),  # collected
                ], continue_on_error=True, context=context)
            except* ZeroDivisionError as e:
                assert len(e.exceptions) == 3
            else:
                pytest.fail("The collected exceptions were not raised")
            assert cancelled.get() == 0

        await asyncio.create_task(_inner(), context=context)



from importlib.metadata import version

# Import submodules only when installed properly
from . import (
    context,
    defer as _defer,
    fork as _fork,
    func,
    iter as _iter,
    server,
    taskgroup,
    timeouts,
    timer,
    utils,
)

__all__ = (
    *context.__all__,
    *_defer.__all__,
    *_fork.__all__,
    *func.__all__,
    *_iter.__all__,
    *server.__all__,
    *taskgroup.__all__,
    *timeouts.__all__,
    *timer.__all__,
    *utils.__all__,
    "__version__",
)

from .context import *     # noqa
from .defer import *       # noqa
from .fork import *        # noqa
from .func import *        # noqa
from .iter import *        # noqa
from .server import *      # noqa
from .taskgroup import *   # noqa
from .timeouts import *    # noqa
from .timer import *       # noqa
from .utils import *       # noqa

__version__ = version("aiotools")



import pytest

import asyncio
import functools
import glob
import logging.config
import multiprocessing as mp
import os
import signal
import sys
import tempfile
import time
from typing import List, Sequence

import aiotools


if os.environ.get('CI', '') and sys.version_info < (3, 9, 0):
    pytest.skip(
        'skipped to prevent kill CI agents due to signals on CI environments',
        allow_module_level=True,
    )


@pytest.fixture
def restore_signal():
    os.setpgrp()
    old_alrm = signal.getsignal(signal.SIGALRM)
    old_intr = signal.getsignal(signal.SIGINT)
    old_term = signal.getsignal(signal.SIGTERM)
    old_intr = signal.getsignal(signal.SIGUSR1)
    yield
    signal.signal(signal.SIGALRM, old_alrm)
    signal.signal(signal.SIGINT, old_intr)
    signal.signal(signal.SIGTERM, old_term)
    signal.signal(signal.SIGUSR1, old_term)


@pytest.fixture
def set_timeout():
    def make_timeout(sec, callback):

        def _callback(signum, frame):
            signal.alarm(0)
            callback()

        signal.signal(signal.SIGALRM, _callback)
        signal.setitimer(signal.ITIMER_REAL, sec)

    yield make_timeout


@pytest.fixture
def exec_recorder():
    f = tempfile.NamedTemporaryFile(
        mode='w', encoding='utf8',
        prefix='aiotools.tests.server.',
    )
    f.close()

    def write(msg: str) -> None:
        path = f"{f.name}.{os.getpid()}"
        with open(path, 'a', encoding='utf8') as writer:
            writer.write(msg + '\n')

    def read() -> Sequence[str]:
        lines: List[str] = []
        for path in glob.glob(f"{f.name}.*"):
            with open(path, 'r', encoding='utf8') as reader:
                lines.extend(line.strip() for line in reader.readlines())
        return lines

    yield write, read

    for path in glob.glob(f"{f.name}.*"):
        os.unlink(path)


def interrupt():
    os.kill(0, signal.SIGINT)


def interrupt_usr1():
    os.kill(os.getpid(), signal.SIGUSR1)


@aiotools.server   # type: ignore
async def myserver_simple(loop, proc_idx, args):
    write = args[0]
    await asyncio.sleep(0)
    write(f'started:{proc_idx}')

    yield

    await asyncio.sleep(0)
    write(f'terminated:{proc_idx}')


def test_server_singleproc(set_timeout, restore_signal, exec_recorder):
    write, read = exec_recorder
    set_timeout(0.2, interrupt)
    aiotools.start_server(
        myserver_simple,
        args=(write,),
    )
    lines = set(read())
    assert 'started:0' in lines
    assert 'terminated:0' in lines


def test_server_multiproc(set_timeout, restore_signal, exec_recorder):
    write, read = exec_recorder
    set_timeout(0.2, interrupt)
    aiotools.start_server(
        myserver_simple,
        num_workers=3,
        args=(write,),
    )
    lines = set(read())
    assert lines == {
        'started:0', 'started:1', 'started:2',
        'terminated:0', 'terminated:1', 'terminated:2',
    }


@aiotools.server  # type: ignore
async def myserver_signal(loop, proc_idx, args):
    write = args[0]
    await asyncio.sleep(0)
    write(f'started:{proc_idx}')

    received_signum = yield

    await asyncio.sleep(0)
    write(f'terminated:{proc_idx}:{received_signum}')


def test_server_multiproc_custom_stop_signals(
    set_timeout,
    restore_signal,
    exec_recorder,
):
    write, read = exec_recorder
    set_timeout(0.2, interrupt_usr1)
    aiotools.start_server(
        myserver_signal,
        num_workers=2,
        stop_signals={signal.SIGUSR1},
        args=(write,),
    )
    lines = set(read())
    assert {'started:0', 'started:1'} < lines
    assert {
        f'terminated:0:{int(signal.SIGUSR1)}',
        f'terminated:1:{int(signal.SIGUSR1)}',
    } < lines


@aiotools.server  # type: ignore
async def myserver_worker_init_error(loop, proc_idx, args):
    write = args[0]

    class _LogAdaptor:
        def __init__(self, writer):
            self.writer = writer

        def write(self, msg):
            msg = msg.strip().replace('\n', ' ')
            self.writer(f'log:{proc_idx}:{msg}')

    log_stream = _LogAdaptor(write)
    logging.config.dictConfig({
        'version': 1,
        'handlers': {
            'console': {
                'class': 'logging.StreamHandler',
                'stream': log_stream,
                'level': 'DEBUG',
            },
        },
        'loggers': {
            'aiotools': {
                'handlers': ['console'],
                'level': 'DEBUG',
            },
        },
    })
    log = logging.getLogger('aiotools')
    write(f'started:{proc_idx}')
    log.debug('hello')
    if proc_idx in (0, 2):
        # delay until other workers start normally.
        await asyncio.sleep(0.1 * proc_idx)
        raise ZeroDivisionError('oops')

    yield

    # should not be reached if errored.
    await asyncio.sleep(0)
    write(f'terminated:{proc_idx}')


def test_server_worker_init_error(restore_signal, exec_recorder):
    write, read = exec_recorder
    aiotools.start_server(
        myserver_worker_init_error,
        num_workers=4,
        args=(write,),
    )
    lines = set(read())
    assert sum(1 if line.startswith('started:') else 0 for line in lines) == 4
    # workers who did not raise errors have already started,
    # and they should have terminated normally
    # when the errorneous worker interrupted the main loop.
    assert sum(1 if line.startswith('terminated:') else 0 for line in lines) == 2
    assert sum(1 if 'hello' in line else 0 for line in lines) == 4
    assert sum(1 if 'ZeroDivisionError: oops' in line else 0 for line in lines) == 2


def test_server_user_main(set_timeout, restore_signal):

    main_enter = False
    main_exit = False

    @aiotools.main
    def mymain_user_main():
        nonlocal main_enter, main_exit
        main_enter = True
        yield 987
        main_exit = True

    @aiotools.server  # type: ignore
    async def myworker_user_main(loop, proc_idx, args):
        assert args[0] == 987  # first arg from user main
        assert args[1] == 123  # second arg from start_server args
        yield

    set_timeout(0.2, interrupt)
    aiotools.start_server(
        myworker_user_main,
        mymain_user_main,
        num_workers=3,
        args=(123,),
    )

    assert main_enter
    assert main_exit


def test_server_user_main_custom_stop_signals(set_timeout, restore_signal):
    main_enter = False
    main_exit = False
    main_signal = None
    worker_signals = mp.Array('i', 3)

    @aiotools.main
    def mymain():
        nonlocal main_enter, main_exit, main_signal
        main_enter = True
        main_signal = yield
        main_exit = True

    @aiotools.server
    async def myworker(loop, proc_idx, args):
        worker_signals = args[0]
        worker_signals[proc_idx] = yield

    def noop(signum, frame):
        pass

    set_timeout(0.2, interrupt_usr1)
    aiotools.start_server(
        myworker,
        mymain,
        num_workers=3,
        stop_signals={signal.SIGUSR1},
        args=(worker_signals,),
    )

    assert main_enter
    assert main_exit
    assert main_signal == signal.SIGUSR1
    assert list(worker_signals) == [signal.SIGUSR1] * 3


def test_server_user_main_tuple(set_timeout, restore_signal):
    main_enter = False
    main_exit = False

    @aiotools.main
    def mymain():
        nonlocal main_enter, main_exit
        main_enter = True
        yield 987, 654
        main_exit = True

    @aiotools.server
    async def myworker(loop, proc_idx, args):
        assert args[0] == 987  # first arg from user main
        assert args[1] == 654  # second arg from user main
        assert args[2] == 123  # third arg from start_server args
        yield

    set_timeout(0.2, interrupt)
    aiotools.start_server(
        myworker,
        mymain,
        num_workers=3,
        args=(123,),
    )

    assert main_enter
    assert main_exit


def test_server_extra_proc(set_timeout, restore_signal):

    extras = mp.Array('i', [0, 0])

    def extra_proc(key, _, pidx, args):
        assert _ is None
        extras[key] = 980 + key
        try:
            while True:
                time.sleep(0.1)
        except KeyboardInterrupt:
            print(f'extra[{key}] interrupted', file=sys.stderr)
        except Exception as e:
            print(f'extra[{key}] exception', e, file=sys.stderr)
        finally:
            print(f'extra[{key}] finish', file=sys.stderr)
            extras[key] = 990 + key

    @aiotools.server
    async def myworker(loop, pidx, args):
        yield

    set_timeout(0.2, interrupt)
    aiotools.start_server(myworker, extra_procs=[
                              functools.partial(extra_proc, 0),
                              functools.partial(extra_proc, 1)],
                          num_workers=3, args=(123, ))

    assert extras[0] == 990
    assert extras[1] == 991


def test_server_extra_proc_custom_stop_signal(set_timeout, restore_signal):

    received_signals = mp.Array('i', [0, 0])

    def extra_proc(key, _, pidx, args):
        received_signals = args[0]
        try:
            while True:
                time.sleep(0.1)
        except aiotools.InterruptedBySignal as e:
            received_signals[key] = e.args[0]

    @aiotools.server
    async def myworker(loop, pidx, args):
        yield

    set_timeout(0.3, interrupt_usr1)
    aiotools.start_server(myworker, extra_procs=[
                              functools.partial(extra_proc, 0),
                              functools.partial(extra_proc, 1)],
                          stop_signals={signal.SIGUSR1},
                          args=(received_signals, ),
                          num_workers=3)

    assert received_signals[0] == signal.SIGUSR1
    assert received_signals[1] == signal.SIGUSR1



from asyncio import events
from asyncio import exceptions
from asyncio import tasks
from typing import Optional

__all__ = ["Supervisor"]


class Supervisor:
    """
    Supervisor is a primitive structure to provide a long-lived context manager scope
    for an indefinite set of subtasks.  During its lifetime, it is free to spawn new
    subtasks at any time.  If the supervisor itself is cancelled from outside or
    :meth:`shutdown()` is called, it will cancel all running tasks immediately, wait
    for their completion, and then exit the context manager block.

    The main difference to :class:`asyncio.TaskGroup` is that it keeps running
    sibling subtasks even when there is an unhandled exception from one of the
    subtasks.

    To prevent memory leaks, a supervisor does not store any result or exception
    from its subtasks.  Instead, the callers must use additional task-done
    callbacks to process subtask results and exceptions.

    Supervisor provides the same analogy to Kotlin's ``SupervisorScope`` and
    Javascript's ``Promise.allSettled()``, while :class:`asyncio.TaskGroup` provides
    the same analogy to Kotlin's ``CoroutineScope`` and Javascript's
    ``Promise.all()``.

    The original implementation is based on DontPanicO's pull request
    (https://github.com/achimnol/cpython/pull/31) and :class:`PersistentTaskGroup`,
    but it is modified *not* to store unhandled subtask exceptions.

    .. versionadded:: 2.0
    """

    def __init__(self):
        self._entered = False
        self._exiting = False
        self._aborting = False
        self._loop = None
        self._parent_task = None
        self._parent_cancel_requested = False
        self._tasks = set()
        self._base_error = None
        self._on_completed_fut = None

    def __repr__(self):
        info = ['']
        if self._tasks:
            info.append(f'tasks={len(self._tasks)}')
        if self._aborting:
            info.append('cancelling')
        elif self._entered:
            info.append('entered')

        info_str = ' '.join(info)
        return f'<Supervisor{info_str}>'

    async def __aenter__(self):
        if self._entered:
            raise RuntimeError(
                f"Supervisor {self!r} has been already entered")
        self._entered = True

        if self._loop is None:
            self._loop = events.get_running_loop()

        self._parent_task = tasks.current_task(self._loop)
        if self._parent_task is None:
            raise RuntimeError(
                f'Supervisor {self!r} cannot determine the parent task')

        return self

    async def __aexit__(self, et, exc, tb):
        self._exiting = True

        if (exc is not None and
                self._is_base_error(exc) and
                self._base_error is None):
            # SystemExit or KeyboardInterrupt in "async with"
            # so we cancel other tasks.
            self._base_error = exc
            self._abort()

        propagate_cancellation_error = \
            exc if et is exceptions.CancelledError else None
        if self._parent_cancel_requested:
            assert self._parent_task is not None
            # If this flag is set we *must* call uncancel().
            if self._parent_task.uncancel() == 0:
                # If there are no pending cancellations left,
                # don't propagate CancelledError.
                propagate_cancellation_error = None

        prop_ex = await self._wait_completion()
        assert not self._tasks
        if prop_ex is not None:
            propagate_cancellation_error = prop_ex

        if self._base_error is not None:
            raise self._base_error

        # Propagate CancelledError if there is one, except if there
        # are other errors -- those have priority.
        if propagate_cancellation_error:
            raise propagate_cancellation_error

        # In the original version, it raises BaseExceptionGroup
        # if there are collected errors in self._errors.
        # This part is deliberately removed to prevent memory leak
        # due to accumulating error objects for an indefinite length of time.

    def create_task(self, coro, *, name=None, context=None):
        if not self._entered:
            raise RuntimeError(f"Supervisor {self!r} has not been entered")
        if self._exiting and not self._tasks:
            raise RuntimeError(f"Supervisor {self!r} is finished")
        if self._aborting:
            raise RuntimeError(f"Supervisor {self!r} is shutting down")
        if context is None:
            task = self._loop.create_task(coro)
        else:
            task = self._loop.create_task(coro, context=context)
        tasks._set_task_name(task, name)
        task.add_done_callback(self._on_task_done)
        self._tasks.add(task)
        return task

    # Since Python 3.8 Tasks propagate all exceptions correctly,
    # except for KeyboardInterrupt and SystemExit which are
    # still considered special.

    def _is_base_error(self, exc: BaseException) -> bool:
        assert isinstance(exc, BaseException)
        return isinstance(exc, (SystemExit, KeyboardInterrupt))

    def _abort(self, msg: Optional[str] = None):
        self._aborting = True

        for t in self._tasks:
            if not t.done():
                t.cancel(msg=msg)

    async def _wait_completion(self):
        # We use while-loop here because "self._on_completed_fut"
        # can be cancelled multiple times if our parent task
        # is being cancelled repeatedly (or even once, when
        # our own cancellation is already in progress)
        propagate_cancellation_error = None
        while self._tasks:
            if self._on_completed_fut is None:
                self._on_completed_fut = self._loop.create_future()

            try:
                await self._on_completed_fut
            except exceptions.CancelledError as ex:
                if not self._aborting:
                    # Our parent task is being cancelled:
                    #
                    #    async def wrapper():
                    #        async with TaskGroup() as g:
                    #            g.create_task(foo)
                    #
                    # "wrapper" is being cancelled while "foo" is
                    # still running.
                    propagate_cancellation_error = ex
                    self._abort(msg=ex.args[0] if ex.args else None)
            self._on_completed_fut = None

        return propagate_cancellation_error

    async def shutdown(self) -> None:
        self._abort(msg="supervisor.shutdown")
        await self._wait_completion()

    def _on_task_done(self, task):
        self._tasks.discard(task)

        if self._on_completed_fut is not None and not self._tasks:
            if not self._on_completed_fut.done():
                self._on_completed_fut.set_result(True)

        if task.cancelled():
            return

        exc = task.exception()
        if exc is None:
            return

        _is_base_error = self._is_base_error(exc)
        if _is_base_error and self._base_error is None:
            self._base_error = exc

        assert self._parent_task is not None
        if self._parent_task.done():
            # Not sure if this case is possible, but we want to handle
            # it anyways.
            self._loop.call_exception_handler({
                'message': f'Task {task!r} has errored out but its parent '
                           f'task {self._parent_task} is already completed',
                'exception': exc,
                'task': task,
            })
            return

        if (
            _is_base_error and
            not self._aborting and
            not self._parent_cancel_requested
        ):
            # For base SystemExit and KeyboardInterrupt ONLY, if parent task
            # *is not* being cancelled, it means that we want to manually cancel
            # it to abort whatever is being run right now in the Supervisor.
            # But we want to mark parent task as "not cancelled" later in __aexit__.
            # Example situation that we need to handle:
            #
            #    async def foo():
            #        try:
            #            async with Supervisor() as s:
            #                s.create_task(crash_soon())
            #                await something  # <- this needs to be waited
            #                                 #    by the Supervisor, unless
            #                                 #    crash_soon() raises either
            #                                 #    SystemExit or KeyboardInterrupt
            #        except Exception:
            #            # Ignore any exceptions raised in the Supervisor
            #            pass
            #        await something_else     # this line has to be called
            #                                 # after TaskGroup is finished.
            self._abort()
            self._parent_cancel_requested = True
            self._parent_task.cancel()



import asyncio
from typing import TypeVar

import pytest

from aiotools import (
    VirtualClock,
)
from aiotools.supervisor import Supervisor

T = TypeVar("T")


async def do_job(delay: float, result: T) -> T:
    await asyncio.sleep(delay)
    return result


async def fail_job(delay: float) -> None:
    await asyncio.sleep(delay)
    1 / 0


@pytest.mark.asyncio
async def test_supervisor_partial_failure():
    results = []
    errors = []
    tasks = []
    with VirtualClock().patch_loop():
        async with Supervisor() as supervisor:
            tasks.append(supervisor.create_task(do_job(0.1, 1)))
            tasks.append(supervisor.create_task(fail_job(0.2)))
            tasks.append(supervisor.create_task(do_job(0.3, 3)))
        for t in tasks:
            try:
                results.append(t.result())
            except Exception as e:
                errors.append(e)
    assert results == [1, 3]
    assert len(errors) == 1
    assert isinstance(errors[0], ZeroDivisionError)


@pytest.mark.asyncio
async def test_supervisor_timeout():
    results = []
    errors = []
    cancelled = 0
    tasks = []
    with VirtualClock().patch_loop():
        with pytest.raises(TimeoutError):
            async with (
                asyncio.timeout(0.25),
                Supervisor() as supervisor,
            ):
                tasks.append(supervisor.create_task(do_job(0.1, 1)))
                tasks.append(supervisor.create_task(fail_job(0.2)))
                tasks.append(supervisor.create_task(do_job(0.3, 3)))
        for t in tasks:
            try:
                results.append(t.result())
            except asyncio.CancelledError:
                cancelled += 1
            except Exception as e:
                errors.append(e)
    assert results == [1]
    assert len(errors) == 1
    assert isinstance(errors[0], ZeroDivisionError)
    assert cancelled == 1



import asyncio
from contextvars import ContextVar, copy_context
from typing import Any, TypeVar

import pytest

from aiotools import (
    aclosing,
    as_completed_safe,
    timeout,
    VirtualClock,
)

T = TypeVar("T")
executed = ContextVar("executed", default=0)
cancelled = ContextVar("cancelled", default=0)


async def do_job(delay: float, result: T) -> T:
    try:
        await asyncio.sleep(delay)
        executed.set(executed.get() + 1)
        return result
    except asyncio.CancelledError:
        await asyncio.sleep(0.1)
        cancelled.set(cancelled.get() + 1)
        raise


async def fail_job(delay: float) -> None:
    try:
        await asyncio.sleep(delay)
        1 / 0
    except asyncio.CancelledError:
        await asyncio.sleep(0.1)
        cancelled.set(cancelled.get() + 1)
        raise


@pytest.mark.asyncio
async def test_as_completed_safe() -> None:
    results = []
    with VirtualClock().patch_loop():
        async with aclosing(as_completed_safe([
            do_job(0.3, 1),
            do_job(0.2, 2),
            do_job(0.1, 3),
        ])) as ag:
            async for result in ag:
                results.append(await result)
    assert results == [3, 2, 1]


@pytest.mark.asyncio
async def test_as_completed_safe_partial_failure() -> None:
    results = []
    errors = []
    with VirtualClock().patch_loop():
        async with aclosing(as_completed_safe([
            do_job(0.1, 1),
            fail_job(0.2),
            do_job(0.3, 3),
            fail_job(0.4),
        ])) as ag:
            async for result in ag:
                try:
                    results.append(await result)
                except Exception as e:
                    errors.append(e)
    assert results == [1, 3]
    assert len(errors) == 2
    assert isinstance(errors[0], ZeroDivisionError)
    assert isinstance(errors[1], ZeroDivisionError)


@pytest.mark.asyncio
async def test_as_completed_safe_immediate_failures() -> None:
    context = copy_context()
    with VirtualClock().patch_loop():

        async def _inner() -> None:
            results: list[Any] = []
            errors: list[Exception] = []
            async with aclosing(as_completed_safe([
                # All these jobs fail at the same tick.
                # Still, we should be able to retrieve all errors.
                fail_job(0),
                fail_job(0),
                fail_job(0),
            ], context=context)) as ag:
                async for result in ag:
                    try:
                        results.append(await result)  # type: ignore
                    except Exception as e:
                        errors.append(e)
            assert results == []
            assert cancelled.get() == 0
            assert len(errors) == 3
            assert isinstance(errors[0], ZeroDivisionError)
            assert isinstance(errors[1], ZeroDivisionError)
            assert isinstance(errors[2], ZeroDivisionError)

        await asyncio.create_task(_inner(), context=context)


@pytest.mark.asyncio
async def test_as_completed_safe_timeout_vanilla() -> None:
    context = copy_context()
    with VirtualClock().patch_loop():

        async def _inner() -> None:
            loop_count = 0
            results = []
            timeouts = 0
            try:
                async with (
                    asyncio.timeout(0.15),
                    aclosing(as_completed_safe([
                        do_job(0.1, 1),
                        # timeout occurs here
                        do_job(0.2, 2),
                        do_job(0.3, 3),
                    ], context=context)) as ag,
                ):
                    async for result in ag:
                        results.append(await result)
                        loop_count += 1
            except asyncio.TimeoutError:
                timeouts += 1

            assert loop_count == 1
            assert executed.get() == 1
            assert cancelled.get() == 2
            assert results == [1]
            assert timeouts == 1

        await asyncio.create_task(_inner(), context=context)


@pytest.mark.asyncio
async def test_as_completed_safe_timeout_in_middle() -> None:
    context = copy_context()
    with VirtualClock().patch_loop():

        async def _inner() -> None:
            loop_count = 0
            results = []
            timeouts = 0
            try:
                async with (
                    asyncio.timeout(0.15),
                    aclosing(as_completed_safe([
                        do_job(0.1, 1),
                        # timeout occurs here
                        do_job(0.2, 2),
                        do_job(0.3, 3),
                    ], context=context)) as ag,
                ):
                    async for result in ag:
                        results.append(await result)
                        await asyncio.sleep(0.1)  # timeout occurs here
                        loop_count += 1
            except asyncio.TimeoutError:
                timeouts += 1

            assert loop_count == 0
            assert executed.get() == 1
            assert cancelled.get() == 2
            assert results == [1]
            assert timeouts == 1

        await asyncio.create_task(_inner(), context=context)


@pytest.mark.asyncio
async def test_as_completed_safe_timeout_custom() -> None:
    context = copy_context()
    with VirtualClock().patch_loop():

        async def _inner() -> None:
            loop_count = 0
            results = []
            timeouts = 0
            try:
                async with (
                    timeout(0.15),
                    aclosing(as_completed_safe([
                        do_job(0.1, 1),
                        # timeout occurs here
                        do_job(0.2, 2),
                        do_job(0.3, 3),
                    ], context=context)) as ag,
                ):
                    async for result in ag:
                        results.append(await result)
                        loop_count += 1
            except asyncio.TimeoutError:
                timeouts += 1

            assert loop_count == 1
            assert executed.get() == 1
            assert cancelled.get() == 2
            assert results == [1]
            assert timeouts == 1

        await asyncio.create_task(_inner(), context=context)


@pytest.mark.asyncio
async def test_as_completed_safe_cancel_from_body():
    context = copy_context()
    with VirtualClock().patch_loop():

        async def _inner() -> None:
            loop_count = 0
            results = []
            with pytest.raises(asyncio.CancelledError):
                async with aclosing(as_completed_safe([
                    do_job(0.1, 1),
                    do_job(0.2, 2),
                    # cancellation occurs here
                    do_job(0.3, 3),
                ], context=context)) as ag:
                    async for result in ag:
                        results.append(await result)
                        loop_count += 1
                        if loop_count == 2:
                            raise asyncio.CancelledError()

            assert loop_count == 2
            assert executed.get() == 2
            assert cancelled.get() == 1
            assert results == [1, 2]

        await asyncio.create_task(_inner(), context=context)


@pytest.mark.asyncio
async def test_as_completed_safe_error_from_body():
    context = copy_context()
    with VirtualClock().patch_loop():

        async def _inner() -> None:
            loop_count = 0
            results = []
            with pytest.raises(ZeroDivisionError):
                async with aclosing(as_completed_safe([
                    do_job(0.1, 1),
                    do_job(0.2, 2),
                    # cancellation occurs here
                    do_job(0.3, 3),
                ], context=context)) as ag:
                    async for result in ag:
                        results.append(await result)
                        loop_count += 1
                        if loop_count == 2:
                            raise ZeroDivisionError()

            assert loop_count == 2
            assert executed.get() == 2
            assert cancelled.get() == 1
            assert results == [1, 2]

        await asyncio.create_task(_inner(), context=context)


@pytest.mark.asyncio
async def test_as_completed_safe_error_from_body_without_aclosing() -> None:
    context = copy_context()
    with VirtualClock().patch_loop():

        async def _inner() -> None:
            loop_count = 0
            results = []
            # This is "unsafe" because it cannot guarantee the completion of
            # the internal supervisor.
            with pytest.raises(ZeroDivisionError):
                async for result in as_completed_safe([
                    do_job(0.1, 1),
                    do_job(0.2, 2),
                    # cancellation occurs here
                    do_job(0.3, 3),
                ], context=context):
                    results.append(await result)
                    loop_count += 1
                    if loop_count == 2:
                        raise ZeroDivisionError()

            assert loop_count == 2
            assert executed.get() == 2
            assert cancelled.get() == 0  # not properly cancelled without aclose()
            assert results == [1, 2]
            # Expected: "Task was destroyed but it is pending!" is observed here.

        await asyncio.create_task(_inner(), context=context)


@pytest.mark.asyncio
async def test_as_completed_safe_error_from_body_aclose_afterwards() -> None:
    context = copy_context()
    with VirtualClock().patch_loop():

        async def _inner() -> None:
            loop_count = 0
            results = []
            ag = as_completed_safe([
                do_job(0.1, 1),
                do_job(0.2, 2),
                # body error occurs here
                do_job(0.3, 3),
            ], context=context)
            try:
                async for result in ag:
                    results.append(await result)
                    loop_count += 1
                    if loop_count == 2:
                        raise ZeroDivisionError()
            except ZeroDivisionError:
                await ag.aclose()
            else:
                pytest.fail("The inner exception should have been propagated out")

            assert loop_count == 2
            assert executed.get() == 2
            assert cancelled.get() == 1
            assert results == [1, 2]

        await asyncio.create_task(_inner(), context=context)



import zmq


if __name__ == '__main__':
    ctx = zmq.Context()
    s = ctx.socket(zmq.PUSH)
    s.connect('tcp://127.0.0.1:5033')

    for _ in range(100):
        s.send(b'hello world')

    s.close()



"""
A set of higher-level coroutine aggregation utilities based on :class:`Supervisor`.
"""

from __future__ import annotations

import asyncio
from contextlib import aclosing
from contextvars import Context
from typing import (
    Any,
    AsyncGenerator,
    Awaitable,
    Iterable,
    List,
    Optional,
    Tuple,
    TypeVar,
    Sequence,
)

from .supervisor import Supervisor

__all__ = (
    "as_completed_safe",
    "gather_safe",
    "race",
)

T = TypeVar("T")


async def as_completed_safe(
    coros: Iterable[Awaitable[T]],
    *,
    context: Optional[Context] = None,
) -> AsyncGenerator[Awaitable[T], None]:
    """
    This is a safer version of :func:`asyncio.as_completed()` which uses
    :class:`aiotools.Supervisor` as an underlying coroutine lifecycle keeper.

    This requires Python 3.11 or higher to work properly with timeouts.

    .. versionadded:: 1.6

    .. versionchanged:: 2.0

       It now uses :class:`aiotools.Supervisor` internally and handles
       timeouts in a bettery way.
    """
    q: asyncio.Queue[asyncio.Task[Any]] = asyncio.Queue()
    remaining = 0

    def result_callback(t: asyncio.Task[Any]) -> None:
        q.put_nowait(t)

    async with Supervisor() as supervisor:
        for coro in coros:
            t = supervisor.create_task(coro, context=context)
            t.add_done_callback(result_callback)
            remaining += 1
        while remaining:
            try:
                t = await q.get()
                remaining -= 1
                try:
                    yield t
                finally:
                    q.task_done()
            except (GeneratorExit, BaseException):
                # GeneratorExit: injected when aclose() is called.
                #                (i.e., the async-for body raises an exception)
                # CancelledError: injected when a timeout occurs
                #                 (i.e., the outer scope cancels the inner)
                # BaseException: injected when the process is going to terminate
                await supervisor.shutdown()
                raise


async def gather_safe(
    coros: Iterable[Awaitable[T]],
    *,
    context: Optional[Context] = None,
) -> List[T | Exception]:
    """
    A safer version of :func:`asyncio.gather()`.  It wraps the passed coroutines
    with a :class:`Supervisor` to ensure the termination of them when returned.

    Additionally, it supports manually setting the context of each subtask.

    Note that if it is cancelled from an outer scope (e.g., timeout), there
    is no way to retrieve partially completed or failed results.
    If you need to process them anyway, you must store the results in a
    separate place in the passed coroutines or use :func:`as_completed_safe()`
    instead.

    .. versionadded:: 2.0
    """
    tasks = []
    async with Supervisor() as supervisor:
        for coro in coros:
            t = supervisor.create_task(coro, context=context)
            tasks.append(t)
        # To ensure safety, the Python version must be 3.7 or higher.
        return await asyncio.gather(*tasks, return_exceptions=True)


async def race(
    coros: Iterable[Awaitable[T]],
    *,
    continue_on_error: bool = False,
    context: Optional[Context] = None,
) -> Tuple[T, Sequence[Exception]]:
    """
    Returns the first result and cancelling all remaining coroutines safely.
    Passing an empty iterable of coroutines is not allowed.

    If ``continue_on_error`` is set False (default), it will raise the first
    exception immediately, cancelling all remaining coroutines.  This behavior is
    same to Javascript's ``Promise.race()``.  The second item of the returned tuple
    is always empty.

    If ``continue_on_error`` is set True, it will keep running until it encounters
    the first successful result.  Then it returns the exceptions as a list in the
    second item of the returned tuple.  If all coroutines fail, it will raise an
    :exc:`ExceptionGroup` to indicate the explicit failure of the entire operation.

    You may use this function to implement a "happy eyeball" algorithm.

    .. versionadded:: 2.0
    """
    async with aclosing(as_completed_safe(coros, context=context)) as ag:
        errors: list[Exception] = []
        async for aresult in ag:
            try:
                result = await aresult
                return result, errors
            except Exception as e:
                if continue_on_error:
                    errors.append(e)
                    continue
                raise
        else:
            if errors:
                raise ExceptionGroup("All coroutines have failed in race()", errors)
            raise ValueError("No coroutines were given to race()")



import asyncio
import sys
import warnings

import aiotools
from aiotools.context import AbstractAsyncContextManager
import pytest


def test_actxmgr_types():

    assert issubclass(aiotools.AsyncContextManager, AbstractAsyncContextManager)

    class boilerplate_ctx():
        async def __aenter__(self):
            return self

        async def __aexit__(self, *exc_info):
            return None

    @aiotools.actxmgr
    async def simple_ctx():
        yield

    async def simple_agen():
        yield

    assert issubclass(boilerplate_ctx, AbstractAsyncContextManager)
    assert isinstance(simple_ctx(), AbstractAsyncContextManager)
    assert not isinstance(simple_agen(), AbstractAsyncContextManager)


@pytest.mark.asyncio
async def test_actxmgr(event_loop):

    step = 0

    @aiotools.actxmgr
    async def simple_ctx(msg):
        nonlocal step
        step = 1
        await asyncio.sleep(0)
        step = 2
        try:
            yield msg
            step = 3
        finally:
            await asyncio.sleep(0)
            step = 4

    step = 0
    async with simple_ctx('hello') as msg:
        assert step == 2
        assert msg == 'hello'
    assert step == 4

    step = 0
    try:
        async with simple_ctx('world') as msg:
            assert step == 2
            assert msg == 'world'
            await asyncio.sleep(0)
            raise ValueError('something wrong')
    except Exception as e:
        await asyncio.sleep(0)
        assert e.args[0] == 'something wrong'
        assert step == 4


@pytest.mark.skipif(sys.version_info >= (3, 7, 0),
                    reason='Deprecated in Python 3.7 or higher')
@pytest.mark.asyncio
async def test_actxmgr_reuse(event_loop):

    @aiotools.actxmgr
    async def simple_ctx(msg):
        yield msg

    cm = simple_ctx('hello')

    async with cm as msg:
        assert msg == 'hello'

    try:
        async with cm as msg:
            assert msg == 'hello'
    except BaseException as exc:
        assert isinstance(exc, RuntimeError)
        assert "didn't yield" in exc.args[0]

    cm = cm._recreate_cm()

    async with cm as msg:
        assert msg == 'hello'


@pytest.mark.asyncio
async def test_actxmgr_exception_in_context_body():

    # Exceptions raised in the context body
    # should be transparently raised.

    @aiotools.actxmgr
    async def simple_ctx(msg):
        await asyncio.sleep(0)
        yield msg
        await asyncio.sleep(0)

    with pytest.raises(ZeroDivisionError):
        async with simple_ctx('hello') as msg:
            assert msg == 'hello'
            raise ZeroDivisionError

    try:
        exc = RuntimeError('oops')
        async with simple_ctx('hello') as msg:
            assert msg == 'hello'
            raise exc
    except BaseException as e:
        assert e is exc
        assert e.args[0] == 'oops'
    else:
        pytest.fail()

    cm = simple_ctx('hello')
    ret = await cm.__aenter__()
    assert ret == 'hello'
    ret = await cm.__aexit__(ValueError, None, None)
    assert not ret


@pytest.mark.asyncio
async def test_actxmgr_exception_in_initialization():

    # Any exception before first yield is just transparently
    # raised out to the context block.

    @aiotools.actxmgr
    async def simple_ctx(msg):
        await asyncio.sleep(0)
        raise ZeroDivisionError
        await asyncio.sleep(0)
        yield msg
        await asyncio.sleep(0)

    try:
        async with simple_ctx('hello') as msg:
            assert msg == 'hello'
    except ZeroDivisionError:
        pass
    else:
        pytest.fail()

    exc = RuntimeError('oops')

    @aiotools.actxmgr
    async def simple_ctx(msg):
        await asyncio.sleep(0)
        raise exc
        await asyncio.sleep(0)
        yield msg
        await asyncio.sleep(0)

    try:
        async with simple_ctx('hello') as msg:
            assert msg == 'hello'
    except BaseException as e:
        assert e is exc
        assert e.args[0] == 'oops'
    else:
        pytest.fail()


@pytest.mark.asyncio
async def test_actxmgr_exception_in_finalization():

    @aiotools.actxmgr
    async def simple_ctx(msg):
        await asyncio.sleep(0)
        yield msg
        await asyncio.sleep(0)
        raise ZeroDivisionError
        await asyncio.sleep(0)

    try:
        async with simple_ctx('hello') as msg:
            assert msg == 'hello'
    except ZeroDivisionError:
        pass
    else:
        pytest.fail()

    exc = RuntimeError('oops')

    @aiotools.actxmgr
    async def simple_ctx(msg):
        yield msg
        raise exc

    try:
        async with simple_ctx('hello') as msg:
            assert msg == 'hello'
    except BaseException as e:
        assert e is exc
        assert e.args[0] == 'oops'
    else:
        pytest.fail()


@pytest.mark.asyncio
async def test_actxmgr_exception_uncaught():

    @aiotools.actxmgr
    async def simple_ctx(msg):
        await asyncio.sleep(0)
        yield msg
        await asyncio.sleep(0)

    try:
        async with simple_ctx('hello'):
            raise IndexError('bomb')
    except BaseException as e:
        assert isinstance(e, IndexError)
        assert e.args[0] == 'bomb'


@pytest.mark.asyncio
async def test_actxmgr_exception_nested():

    @aiotools.actxmgr
    async def simple_ctx(msg):
        yield msg

    try:
        async with simple_ctx('hello') as msg1:
            async with simple_ctx('world') as msg2:
                assert msg1 == 'hello'
                assert msg2 == 'world'
                raise IndexError('bomb1')
    except BaseException as exc:
        assert isinstance(exc, IndexError)
        assert 'bomb1' == exc.args[0]
    else:
        pytest.fail()


@pytest.mark.asyncio
async def test_actxmgr_exception_chained():

    @aiotools.actxmgr
    async def simple_ctx(msg):
        try:
            await asyncio.sleep(0)
            yield msg
        except Exception as e:
            await asyncio.sleep(0)
            # exception is chained
            raise ValueError('bomb2') from e

    try:
        async with simple_ctx('hello') as msg:
            assert msg == 'hello'
            raise IndexError('bomb1')
    except BaseException as exc:
        assert isinstance(exc, ValueError)
        assert 'bomb2' == exc.args[0]
        assert isinstance(exc.__cause__, IndexError)
        assert 'bomb1' == exc.__cause__.args[0]
    else:
        pytest.fail()


@pytest.mark.asyncio
async def test_actxmgr_exception_replaced():

    @aiotools.actxmgr
    async def simple_ctx(msg):
        try:
            await asyncio.sleep(0)
            yield msg
        except Exception:
            await asyncio.sleep(0)
            # exception is replaced
            raise ValueError('bomb2')

    try:
        async with simple_ctx('hello') as msg:
            assert msg == 'hello'
            raise IndexError('bomb1')
    except BaseException as exc:
        assert isinstance(exc, ValueError)
        assert 'bomb2' == exc.args[0]
        assert exc.__cause__ is None
    else:
        pytest.fail()


@pytest.mark.asyncio
async def test_actxmgr_stopaiter(event_loop):

    @aiotools.actxmgr
    async def simple_ctx():
        await asyncio.sleep(0)
        yield 1
        await asyncio.sleep(0)
        yield 2
        await asyncio.sleep(0)

    cm = simple_ctx()
    ret = await cm.__aenter__()
    assert ret == 1
    ret = await cm.__aexit__(StopAsyncIteration, None, None)
    assert ret is False

    step = 0

    @aiotools.actxmgr
    async def simple_ctx():
        nonlocal step
        step = 1
        try:
            await asyncio.sleep(0)
            yield
        finally:
            step = 2
            await asyncio.sleep(0)
            raise StopAsyncIteration('x')

    try:
        step = 0
        async with simple_ctx():
            assert step == 1
    except RuntimeError:
        # converted to RuntimeError
        assert step == 2
    else:
        pytest.fail()


@pytest.mark.asyncio
async def test_actxmgr_transparency(event_loop):

    step = 0

    @aiotools.actxmgr
    async def simple_ctx():
        nonlocal step
        step = 1
        await asyncio.sleep(0)
        yield
        step = 2
        await asyncio.sleep(0)

    try:
        exc = StopAsyncIteration('x')
        step = 0
        async with simple_ctx():
            assert step == 1
            raise exc
    except StopAsyncIteration as e:
        # exception is not handled
        assert e is exc
        assert step == 1
    else:
        pytest.fail()

    @aiotools.actxmgr
    async def simple_ctx():
        nonlocal step
        step = 1
        await asyncio.sleep(0)
        yield
        step = 2
        await asyncio.sleep(0)

    try:
        step = 0
        async with simple_ctx():
            assert step == 1
            raise ValueError
    except ValueError:
        # exception is not handled
        assert step == 1
    else:
        pytest.fail()

    @aiotools.actxmgr
    async def simple_ctx():
        nonlocal step
        step = 1
        await asyncio.sleep(0)
        try:
            yield
        finally:
            step = 2
            await asyncio.sleep(0)

    try:
        exc = StopAsyncIteration('x')
        step = 0
        async with simple_ctx():
            assert step == 1
            raise exc
    except StopAsyncIteration as e:
        # exception is not handled,
        # but context is finalized
        assert e is exc
        assert step == 2
    else:
        pytest.fail()

    @aiotools.actxmgr
    async def simple_ctx():
        nonlocal step
        step = 1
        await asyncio.sleep(0)
        try:
            yield
        finally:
            step = 2
            await asyncio.sleep(0)

    try:
        step = 0
        async with simple_ctx():
            assert step == 1
            raise ValueError
    except ValueError:
        # exception is not handled,
        # but context is finalized
        assert step == 2
    else:
        pytest.fail()


@pytest.mark.asyncio
async def test_actxmgr_no_stop(event_loop):

    @aiotools.actxmgr
    async def simple_ctx(msg):
        await asyncio.sleep(0)
        yield msg
        await asyncio.sleep(0)
        yield msg
        await asyncio.sleep(0)

    try:
        async with simple_ctx('hello') as msg:
            assert msg == 'hello'
    except RuntimeError as exc:
        assert "didn't stop" in exc.args[0]
    else:
        pytest.fail()

    try:
        async with simple_ctx('hello') as msg:
            assert msg == 'hello'
            raise ValueError('oops')
    except ValueError as exc:
        assert exc.args[0] == 'oops'
    else:
        pytest.fail()

    @aiotools.actxmgr
    async def simple_ctx(msg):
        try:
            await asyncio.sleep(0)
            yield msg
        finally:
            await asyncio.sleep(0)
            yield msg

    try:
        async with simple_ctx('hello') as msg:
            assert msg == 'hello'
            raise ValueError('oops')
    except RuntimeError as exc:
        assert "didn't stop after" in exc.args[0]
    else:
        pytest.fail()


@pytest.mark.asyncio
async def test_actxmgr_no_yield(event_loop):

    @aiotools.actxmgr
    async def no_yield_ctx1(msg):
        pass

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        try:
            async with no_yield_ctx1('hello'):
                pass
        except RuntimeError as exc:
            assert "must be an async-gen" in exc.args[0]
        except AttributeError:  # in Python 3.7
            pass
        except TypeError as exc:  # in Python 3.10
            assert "not an async iterator" in exc.args[0]
        else:
            pytest.fail()


@pytest.mark.skipif(sys.version_info >= (3, 7, 0),
                    reason='Deprecated in Python 3.7 or higher')
@pytest.mark.asyncio
async def test_actxdecorator(event_loop):

    step = 0

    class myacontext(aiotools.AsyncContextDecorator):

        async def __aenter__(self):
            nonlocal step
            step = 1
            return self

        async def __aexit__(self, exc_type, exc, tb):
            nonlocal step
            step = 2
            return False

    # NOTE: don't forget the ending parenthesis which instantiates myacontext
    @myacontext()
    async def myfunc():
        assert step == 1

    assert step == 0
    await myfunc()
    assert step == 2

    @myacontext()
    async def myfunc():
        assert step == 1
        raise RuntimeError('oops')

    step = 0
    assert step == 0
    try:
        await myfunc()
    except BaseException as e:
        assert e.args[0] == 'oops'
        assert step == 2
    finally:
        assert step == 2

    @myacontext()
    async def myfunc():
        assert step == 1

    # Use as a plain async context manager.
    step = 0
    assert step == 0
    async with myacontext():
        assert step == 1
    assert step == 2


@pytest.mark.asyncio
async def test_actxgroup(event_loop):

    # Test basic function.
    exit_count = 0

    @aiotools.actxmgr
    async def ctx(a):
        nonlocal exit_count
        await asyncio.sleep(0)
        yield a + 10
        await asyncio.sleep(0)
        exit_count += 1

    ctxgrp = aiotools.actxgroup()

    for i in range(3):
        ctxgrp.add(ctx(i))

    async with ctxgrp as values:
        assert values[0] == 10
        assert values[1] == 11
        assert values[2] == 12
        assert len(ctxgrp._cm_yields) == 3

    assert exit_count == 3
    assert len(ctxgrp._cm_yields) == 0
    returns = ctxgrp.exit_states()
    assert not returns[0]
    assert not returns[1]
    assert not returns[2]

    # Test generator/iterator initialization
    exit_count = 0
    ctxgrp = aiotools.actxgroup(ctx(i) for i in range(3))

    async with ctxgrp as values:
        assert values[0] == 10
        assert values[1] == 11
        assert values[2] == 12
        assert len(ctxgrp._cm_yields) == 3

    assert exit_count == 3
    assert len(ctxgrp._cm_yields) == 0
    returns = ctxgrp.exit_states()
    assert not returns[0]
    assert not returns[1]
    assert not returns[2]


@pytest.mark.asyncio
async def test_actxgroup_exception_from_cm(event_loop):

    @aiotools.actxmgr
    async def ctx1(a):
        await asyncio.sleep(0)
        raise asyncio.CancelledError
        await asyncio.sleep(0)
        yield a

    @aiotools.actxmgr
    async def ctx2(a):
        await asyncio.sleep(0)
        raise ZeroDivisionError
        await asyncio.sleep(0)
        yield a

    ctxgrp = aiotools.actxgroup([ctx1(1), ctx2(2)])

    async with ctxgrp as values:
        assert isinstance(values[0], asyncio.CancelledError)
        assert isinstance(values[1], ZeroDivisionError)

    @aiotools.actxmgr
    async def ctx3(a):
        yield a
        raise asyncio.CancelledError

    @aiotools.actxmgr
    async def ctx4(a):
        yield a
        raise ZeroDivisionError

    ctxgrp = aiotools.actxgroup([ctx3(1), ctx4(2)])

    async with ctxgrp as values:
        assert values[0] == 1
        assert values[1] == 2

    exits = ctxgrp.exit_states()
    assert isinstance(exits[0], asyncio.CancelledError)
    assert isinstance(exits[1], ZeroDivisionError)


@pytest.mark.asyncio
async def test_actxgroup_exception_from_body(event_loop):

    exit_count = 0

    @aiotools.actxmgr
    async def ctx(a):
        nonlocal exit_count
        await asyncio.sleep(0)
        yield a
        await asyncio.sleep(0)
        # yield raises the exception from the context body.
        # If not handled, finalization will not be executed.
        exit_count += 1

    ctxgrp = aiotools.actxgroup([ctx(1), ctx(2)])

    try:
        async with ctxgrp as values:
            assert values[0] == 1
            assert values[1] == 2
            raise ZeroDivisionError
    except Exception as e:
        assert isinstance(e, ZeroDivisionError)

    exits = ctxgrp.exit_states()
    assert not exits[0]  # __aexit__ are called successfully
    assert not exits[1]
    assert exit_count == 0   # but they errored internally

    exit_count = 0

    @aiotools.actxmgr
    async def ctx(a):
        nonlocal exit_count
        try:
            await asyncio.sleep(0)
            yield a
        finally:
            await asyncio.sleep(0)
            # Ensure finalization is executed.
            exit_count += 1

    ctxgrp = aiotools.actxgroup([ctx(1), ctx(2)])

    try:
        async with ctxgrp as values:
            assert values[0] == 1
            assert values[1] == 2
            raise ZeroDivisionError
    except Exception as e:
        assert isinstance(e, ZeroDivisionError)

    exits = ctxgrp.exit_states()
    assert not exits[0]  # __aexit__ are called successfully
    assert not exits[1]
    assert exit_count == 2   # they also suceeeded


@pytest.mark.asyncio
async def test_aclosing(event_loop):

    finalized = False

    async def myiter():
        nonlocal finalized
        try:
            yield 1
            await asyncio.sleep(0.2)
            yield 10
            await asyncio.sleep(0.2)
            yield 100
        except asyncio.CancelledError:
            await asyncio.sleep(0.2)
            finalized = True
            raise

    mysum = 0
    detect_cancellation = False

    async def mytask():
        nonlocal mysum, detect_cancellation
        try:
            async with aiotools.aclosing(myiter()) as agen:
                async for x in agen:
                    mysum += x
        except asyncio.CancelledError:
            detect_cancellation = True

    t = event_loop.create_task(mytask())
    await asyncio.sleep(0.3)
    t.cancel()
    await t

    assert mysum == 11
    assert finalized
    assert detect_cancellation



"""
Based on :doc:`aiotools.context`, this module provides an automated lifecycle
management for multi-process servers with explicit initialization steps and
graceful shutdown steps.

.. function:: server(func)

    A decorator wrapper for :class:`AsyncServerContextManager`.

    Usage example:

    .. code:: python

       @aiotools.server
       async def myserver(loop, pidx, args):
           await do_init(args)
           stop_sig = yield
           if stop_sig == signal.SIGINT:
               await do_graceful_shutdown()
           else:
               await do_forced_shutdown()

       aiotools.start_server(myserver, ...)
"""

import asyncio
import functools
import inspect
import logging
import os
import signal
import struct
import sys
import threading
from contextlib import (
    AbstractContextManager, ContextDecorator,
)
from contextvars import ContextVar
from typing import (
    Any,
    Callable,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
)

from .compat import all_tasks, current_task, get_running_loop
from .context import AbstractAsyncContextManager
from .fork import AbstractChildProcess, afork, _has_pidfd

__all__ = (
    'main',
    'start_server',
    'process_index',
    'AsyncServerContextManager',
    'ServerMainContextManager',
    'InterruptedBySignal',
)

log = logging.getLogger(__name__)

process_index: ContextVar[int] = ContextVar('process_index')


class InterruptedBySignal(BaseException):
    """
    A new :class:`BaseException` that represents interruption by an arbitrary UNIX
    signal.

    Since this is a :class:`BaseException` instead of :class:`Exception`, it behaves
    like :class:`KeyboardInterrupt` and :class:`SystemExit` exceptions (i.e.,
    bypassing except clauses catching the :class:`Exception` type only)

    The first argument of this exception is the signal number received.
    """
    pass


class AsyncServerContextManager(AbstractAsyncContextManager):
    """
    A modified version of :func:`contextlib.asynccontextmanager`.

    The implementation detail is mostly taken from the ``contextlib`` standard
    library, with a minor change to inject ``self.yield_return`` into the wrapped
    async generator.
    """

    yield_return: Optional[signal.Signals]

    def __init__(self, func: Callable[..., Any], args, kwargs):
        if not inspect.isasyncgenfunction(func):
            raise RuntimeError('Context manager function must be '
                               'an async-generator')
        self._agen = func(*args, **kwargs)
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.yield_return = None

    async def __aenter__(self):
        try:
            return (await self._agen.__anext__())
        except StopAsyncIteration:
            raise RuntimeError("async-generator didn't yield") from None

    async def __aexit__(self, exc_type, exc_value, tb):
        if exc_type is None:
            try:
                # Here is the modified part.
                await self._agen.asend(self.yield_return)
            except StopAsyncIteration:
                return
            else:
                raise RuntimeError("async-generator didn't stop") from None
        else:
            if exc_value is None:
                exc_value = exc_type()
            try:
                await self._agen.athrow(exc_type, exc_value, tb)
                raise RuntimeError("async-generator didn't stop after athrow()")
            except StopAsyncIteration as exc_new_value:
                return exc_new_value is not exc_value
            except RuntimeError as exc_new_value:
                if exc_new_value is exc_value:
                    return False
                if isinstance(exc_value, (StopIteration, StopAsyncIteration)):
                    if exc_new_value.__cause__ is exc_value:
                        return False
                raise
            except (BaseException, asyncio.CancelledError) as exc:
                if exc is not exc_value:
                    raise


class ServerMainContextManager(AbstractContextManager,
                               ContextDecorator):
    """
    A modified version of :func:`contextlib.contextmanager`.

    The implementation detail is mostly taken from the ``contextlib`` standard
    library, with a minor change to inject ``self.yield_return`` into the wrapped
    generator.
    """

    yield_return: Optional[signal.Signals]

    def __init__(self, func, args, kwargs):
        self.gen = func(*args, **kwargs)
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.yield_return = None

    def __enter__(self):
        del self.args, self.kwargs, self.func
        try:
            return next(self.gen)
        except StopIteration:
            raise RuntimeError("generator didn't yield") from None

    def __exit__(self, type, value, traceback):
        if type is None:
            try:
                self.gen.send(self.yield_return)
            except StopIteration:
                return False
            else:
                raise RuntimeError("generator didn't stop")
        else:
            if value is None:
                value = type()
            try:
                self.gen.throw(type, value, traceback)
            except StopIteration as exc:
                return exc is not value
            except RuntimeError as exc:
                if exc is value:
                    return False
                if type is StopIteration and exc.__cause__ is value:
                    return False
                raise
            except:  # noqa
                if sys.exc_info()[1] is value:
                    return False
                raise
            raise RuntimeError("generator didn't stop after throw()")


# This is a dirty hack to implement "module callable".
# NOTE: only works in Python 3.5 or higher.

def _server_ctxmgr(func):
    @functools.wraps(func)
    def helper(*args, **kwargs):
        return AsyncServerContextManager(func, args, kwargs)
    return helper


class _ServerModule(sys.modules[__name__].__class__):  # type: ignore
    def __call__(self, func):
        return _server_ctxmgr(func)


sys.modules[__name__].__class__ = _ServerModule


def _main_ctxmgr(func):
    """
    A decorator wrapper for :class:`ServerMainContextManager`

    Usage example:

    .. code:: python

       @aiotools.main
       def mymain():
           server_args = do_init()
           stop_sig = yield server_args
           if stop_sig == signal.SIGINT:
               do_graceful_shutdown()
           else:
               do_forced_shutdown()

       aiotools.start_server(..., main_ctxmgr=mymain, ...)
    """
    @functools.wraps(func)
    def helper(*args, **kwargs):
        return ServerMainContextManager(func, args, kwargs)
    return helper


main = _main_ctxmgr


def setup_child_watcher(loop: asyncio.AbstractEventLoop) -> None:
    try:
        watcher_cls = getattr(asyncio, 'PidfdChildWatcher', None)
        if _has_pidfd and watcher_cls:
            watcher = watcher_cls()
            asyncio.set_child_watcher(watcher)
        else:
            # Just get the default child watcher.
            watcher = asyncio.get_child_watcher()
        if not watcher.is_active():
            watcher.attach_loop(loop)
    except NotImplementedError:
        pass  # for uvloop


async def cancel_all_tasks() -> None:
    loop = get_running_loop()
    cancelled_tasks = []
    for task in all_tasks():
        if not task.done() and task is not current_task():
            task.cancel()
            cancelled_tasks.append(task)
    await asyncio.gather(*cancelled_tasks, return_exceptions=True)
    for task in cancelled_tasks:
        if task.cancelled():
            continue
        if task.exception() is not None:
            loop.call_exception_handler({
                'message': 'unhandled exception during loop shutdown',
                'exception': task.exception(),
                'task': task,
            })


def _worker_main(
    worker_actxmgr: Callable[
        [asyncio.AbstractEventLoop, int, Sequence[Any]],
        AsyncServerContextManager,
    ],
    stop_signals: Set[signal.Signals],
    intr_pipe_wfd: int,
    proc_idx: int,
    args: Sequence[Any],
) -> int:
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    setup_child_watcher(loop)
    interrupted = asyncio.Event()
    ctx = worker_actxmgr(loop, proc_idx, args)
    process_index.set(proc_idx)
    forever_future = loop.create_future()

    def handle_stop_signal(signum):
        if interrupted.is_set():
            pass
        else:
            interrupted.set()
            ctx.yield_return = signum
            forever_future.cancel()

    for signum in stop_signals:
        loop.add_signal_handler(
            signum,
            functools.partial(handle_stop_signal, signum),
        )
    # Allow the worker to be interrupted during initialization
    # (in case of initialization failures in other workers)
    signal.pthread_sigmask(signal.SIG_UNBLOCK, stop_signals)

    async def _wrapped_worker():
        err_ctx = 'enter'
        try:
            async with ctx:
                err_ctx = 'body'
                try:
                    await forever_future
                except asyncio.CancelledError:
                    pass
                finally:
                    err_ctx = 'exit'
        except Exception:
            if err_ctx != 'body':
                err_ctx_str = 'initialization' if err_ctx == 'enter' else 'shutdown'
                log.exception(f'Worker {proc_idx}: '
                              f'Error during context manager {err_ctx_str}')
                os.write(intr_pipe_wfd, struct.pack('i', proc_idx))
            raise

    try:
        loop.run_until_complete(_wrapped_worker())
    finally:
        try:
            loop.run_until_complete(cancel_all_tasks())
            loop.run_until_complete(loop.shutdown_asyncgens())
            try:
                loop.run_until_complete(loop.shutdown_default_executor())
            except (AttributeError, NotImplementedError):  # for uvloop
                pass
        finally:
            loop.close()
        return 0


def _extra_main(
    extra_func,
    stop_signals: Set[signal.Signals],
    proc_idx: int,
    args: Sequence[Any],
) -> int:
    interrupted = threading.Event()
    process_index.set(proc_idx)

    # Since signals only work for the main thread in Python,
    # extra processes in use_threading=True mode should check
    # the intr_event by themselves (probably in their loops).

    def raise_stop(signum, frame):
        if interrupted.is_set():
            pass
        else:
            interrupted.set()
            if signum == signal.SIGINT:
                raise KeyboardInterrupt
            elif signum == signal.SIGTERM:
                raise SystemExit
            else:
                raise InterruptedBySignal(signum)

    # restore signal handler.
    for signum in stop_signals:
        signal.signal(signum, raise_stop)
    signal.pthread_sigmask(signal.SIG_UNBLOCK, stop_signals)
    intr_event = None

    try:
        if not interrupted.is_set():
            extra_func(intr_event, proc_idx, args)
    except (SystemExit, KeyboardInterrupt, InterruptedBySignal):
        log.warning(f'extra_proc[{proc_idx}] did not handle stop signals.')
    finally:
        # same as in _worker_main()
        signal.pthread_sigmask(signal.SIG_BLOCK, stop_signals)
        return 0


def start_server(
    worker_actxmgr: Callable[
        [asyncio.AbstractEventLoop, int, Sequence[Any]],
        AsyncServerContextManager],
    main_ctxmgr: Optional[Callable[[], ServerMainContextManager]] = None,
    extra_procs: Iterable[Callable] = tuple(),
    stop_signals: Iterable[signal.Signals] = (
        signal.SIGINT,
        signal.SIGTERM
    ),
    num_workers: int = 1,
    args: Iterable[Any] = tuple(),
    wait_timeout: Optional[float] = None,
) -> None:
    """
    Starts a multi-process server where each process has their own individual
    asyncio event loop.  Their lifecycles are automantically managed -- if the
    main program receives one of the signals specified in ``stop_signals`` it
    will initiate the shutdown routines on each worker that stops the event
    loop gracefully.

    Args:
        worker_actxmgr: An asynchronous context manager that dicates the
                        initialization and shutdown steps of each worker.
                        It should accept the following three arguments:

                        * **loop**: the asyncio event loop created and set
                          by aiotools
                        * **pidx**: the 0-based index of the worker
                          (use this for per-worker logging)
                        * **args**: a concatenated tuple of values yielded by
                          **main_ctxmgr** and the user-defined arguments in
                          **args**.

                        aiotools automatically installs an interruption handler
                        that calls ``loop.stop()`` to the given event loop,
                        regardless of using either threading or
                        multiprocessing.

        main_ctxmgr: An optional context manager that performs global
                     initialization and shutdown steps of the whole program.
                     It may yield one or more values to be passed to worker
                     processes along with **args** passed to this function.
                     There is no arguments passed to those functions since
                     you can directly access ``sys.argv`` to parse command
                     line arguments and/or read user configurations.

        extra_procs: An iterable of functions that consist of extra processes
                     whose lifecycles are synchronized with other workers.
                     They should set up their own signal handlers.

                     It should accept the following three arguments:

                     * **intr_event**: Always ``None``, kept for legacy
                     * **pidx**: same to **worker_actxmgr** argument
                     * **args**: same to **worker_actxmgr** argument

        stop_signals: A list of UNIX signals that the main program to
                      recognize as termination signals.

        num_workers: The number of children workers.

        args: The user-defined arguments passed to workers and extra
              processes.  If **main_ctxmgr** yields one or more values,
              they are *prepended* to this user arguments when passed to
              workers and extra processes.

        wait_timeout: The timeout in seconds before forcibly killing all
                      remaining child processes after sending initial stop
                      signals.

    Returns:
        None

    .. versionchanged:: 0.3.2

       The name of argument **num_proc** is changed to **num_workers**.
       Even if **num_workers** is 1, a child is created instead of
       doing everything at the main thread.

    .. versionadded:: 0.3.2

       The argument ``extra_procs`` and ``main_ctxmgr``.

    .. versionadded:: 0.4.0

       Now supports use of threading instead of multiprocessing via
       **use_threading** option.

    .. versionchanged:: 0.8.0

       Now **worker_actxmgr** must be an instance of
       :class:`AsyncServerContextManager` or async generators decorated by
       ``@aiotools.server``.

       Now **main_ctxmgr** must be an instance of :class:`ServerMainContextManager`
       or plain generators decorated by ``@aiotools.main``.

       The usage is same to asynchronous context managers, but optionally you can
       distinguish the received stop signal by retrieving the return value of the
       ``yield`` statement.

       In **extra_procs** in non-threaded mode, stop signals are converted into
       either one of :class:`KeyboardInterrupt`, :class:`SystemExit`, or
       :class:`InterruptedBySignal` exception.

    .. versionadded:: 0.8.4

       **start_method** argument can be set to change the subprocess spawning
       implementation.

    .. deprecated:: 1.2.0

       The **start_method** and **use_threading** arguments, in favor of our new
       :func:`afork()` function which provides better synchronization and pid-fd
       support.

    .. versionchanged:: 1.2.0

       The **extra_procs** will be always separate processes since **use_threading**
       is deprecated and thus **intr_event** arguments are now always ``None``.

    .. versionadded:: 1.5.5

       The **wait_timeout** argument.
    """

    @_main_ctxmgr
    def noop_main_ctxmgr():
        yield

    assert stop_signals

    if hasattr(asyncio, 'get_running_loop'):
        # only for Python 3.7+
        try:
            asyncio.get_running_loop()
        except RuntimeError:
            # there should be none.
            pass
        else:
            raise RuntimeError(
                'aiotools.start_server() cannot be called inside '
                'a running event loop.')

    if main_ctxmgr is None:
        main_ctxmgr = noop_main_ctxmgr

    children: List[AbstractChildProcess] = []
    sigblock_mask = frozenset(stop_signals)

    main_ctx = main_ctxmgr()

    # temporarily block signals and register signal handlers to main_loop
    signal.pthread_sigmask(signal.SIG_BLOCK, sigblock_mask)

    main_loop = asyncio.new_event_loop()
    asyncio.set_event_loop(main_loop)
    main_future = main_loop.create_future()

    # to make subprocess working in child threads
    setup_child_watcher(main_loop)

    # build a main-to-worker interrupt channel using signals
    def handle_stop_signal(signum: signal.Signals) -> None:
        main_ctx.yield_return = signum  # type: ignore
        for child in children:
            child.send_signal(signum)
        # instead of main_loop.stop(), we use a future here
        # NOT to interrupt the cleanup routines in
        # an arbitrary timing upon async child failures.
        main_future.cancel()

    for signum in stop_signals:
        main_loop.add_signal_handler(
            signum,
            functools.partial(handle_stop_signal, signum),
        )

    # build a reliable worker-to-main interrupt channel using a pipe
    # (workers have no idea whether the main interrupt is enabled/disabled)
    def handle_child_interrupt(fd: int) -> None:
        child_idx = struct.unpack('i', os.read(fd, 4))[0]  # noqa
        log.debug(f'Child {child_idx} has interrupted the main program.')
        # self-interrupt to initiate the main-to-worker interrupts
        signal.pthread_sigmask(signal.SIG_UNBLOCK, {signal.SIGINT})
        os.kill(0, signal.SIGINT)

    child_intr_pipe: Tuple[int, int] = os.pipe()
    rfd = child_intr_pipe[0]
    main_loop.add_reader(rfd, handle_child_interrupt, rfd)

    # start
    with main_ctx as main_args:

        # retrieve args generated by the user-defined main
        if main_args is None:
            main_args = tuple()
        if not isinstance(main_args, tuple):
            main_args = (main_args, )

        # spawn managed async workers
        for i in range(num_workers):
            try:
                p = main_loop.run_until_complete(afork(functools.partial(
                    _worker_main,
                    worker_actxmgr,
                    stop_signals,
                    child_intr_pipe[1],
                    i,
                    main_args + args,
                )))
            except RuntimeError as e:
                if 'loop stopped' in e.args[0]:
                    log.warning(
                        "skipping spawning of child[worker]:%d due to "
                        "async failure(s) of other child process",
                        i,
                    )
                    continue
                raise
            children.append(p)

        # spawn extra workers
        for i, f in enumerate(extra_procs):
            try:
                p = main_loop.run_until_complete(afork(functools.partial(
                    _extra_main,
                    f,
                    stop_signals,
                    num_workers + i,
                    main_args + args,
                )))
            except RuntimeError as e:
                if 'loop stopped' in e.args[0]:
                    log.warning(
                        "skipping spawning of child[extra]:%d due to "
                        "async failure(s) of other child process",
                        num_workers + i,
                    )
                    continue
                raise
            children.append(p)
        try:
            # unblock the stop signals for user/external interrupts.
            signal.pthread_sigmask(signal.SIG_UNBLOCK, sigblock_mask)

            # run!
            try:
                main_loop.run_until_complete(main_future)
            except asyncio.CancelledError:
                pass
            finally:
                # if interrupted, wait for workers to finish.
                try:
                    main_loop.run_until_complete(
                        asyncio.wait_for(asyncio.gather(
                            *[child.wait() for child in children],
                            return_exceptions=True,
                        ), wait_timeout)
                    )
                except asyncio.TimeoutError:
                    log.warning(
                        "Timeout during waiting for child processes; "
                        "killing all",
                    )
                    for child in children:
                        child.send_signal(signal.SIGKILL)
        finally:
            try:
                main_loop.run_until_complete(cancel_all_tasks())
                main_loop.run_until_complete(main_loop.shutdown_asyncgens())
                try:
                    main_loop.run_until_complete(
                        main_loop.shutdown_default_executor()
                    )
                except (AttributeError, NotImplementedError):  # for uvloop
                    pass
            finally:
                main_loop.close()
                asyncio.set_event_loop(None)



import asyncio

import aiotools


async def echo(reader, writer):
    data = await reader.read(100)
    writer.write(data)
    await writer.drain()
    writer.close()


@aiotools.actxmgr
async def worker_main(loop, pidx, args):

    # Create a listening socket with SO_REUSEPORT option so that each worker
    # process can share the same listening port and the kernel balances
    # incoming connections across multiple worker processes.
    server = await asyncio.start_server(echo, '0.0.0.0', 8888,
        reuse_port=True, loop=loop)
    print(f'[{pidx}] started')

    yield  # wait until terminated

    server.close()
    await server.wait_closed()
    print(f'[{pidx}] terminated')


if __name__ == '__main__':
    # Run the above server using 4 worker processes.
    aiotools.start_server(worker_main, num_workers=4)



import asyncio

import pytest

from aiotools.func import apartial, lru_cache


async def do(a, b, *, c=1, d=2):
    '''hello world'''
    return (a, b, c, d)


@pytest.mark.asyncio
async def test_apartial_orig():
    do2 = apartial(do)
    ret = await do2(1, 2, c=3, d=4)
    assert ret == (1, 2, 3, 4)


@pytest.mark.asyncio
async def test_apartial_args():
    do2 = apartial(do, 9)
    ret = await do2(2, c=5, d=6)
    assert ret == (9, 2, 5, 6)


@pytest.mark.asyncio
async def test_apartial_kwargs():
    do2 = apartial(do, c=8)
    ret = await do2(1, 2, d=4)
    assert ret == (1, 2, 8, 4)


@pytest.mark.asyncio
async def test_apartial_args_kwargs():
    do2 = apartial(do, 9, c=8)
    ret = await do2(7, d=6)
    assert ret == (9, 7, 8, 6)


@pytest.mark.asyncio
async def test_apartial_wraps():
    do2 = apartial(do)
    assert do2.__doc__.strip() == 'hello world'
    assert do2.__doc__ == do.__doc__
    assert do.__name__ == 'do'
    assert do2.__name__ == 'do'


@pytest.mark.asyncio
async def test_lru_cache():

    calc_count = 0

    @lru_cache(maxsize=2)
    async def calc(n):
        '''testing'''
        nonlocal calc_count
        await asyncio.sleep(0)
        calc_count += 1
        return n * n

    assert calc.__name__ == 'calc'
    assert calc.__doc__ == 'testing'

    assert (await calc(1)) == 1
    assert calc_count == 1
    assert (await calc(2)) == 4
    assert calc_count == 2
    assert (await calc(1)) == 1
    assert calc_count == 2
    assert (await calc(3)) == 9
    assert calc_count == 3
    assert (await calc(1)) == 1  # evicted and re-executed
    assert calc_count == 4
    assert (await calc(1)) == 1  # cached again
    assert calc_count == 4

    with pytest.raises(NotImplementedError):
        calc.cache_info()

    calc.cache_clear()

    assert (await calc(1)) == 1
    assert calc_count == 5
    assert (await calc(3)) == 9
    assert calc_count == 6


@pytest.mark.asyncio
async def test_lru_cache_with_expiration():

    calc_count = 0

    @lru_cache(maxsize=2)
    async def calc_no_exp(n):
        nonlocal calc_count
        await asyncio.sleep(0)
        calc_count += 1
        return n * n

    assert (await calc_no_exp(3)) == 9
    assert calc_count == 1
    assert (await calc_no_exp(3)) == 9
    assert calc_count == 1
    await asyncio.sleep(0.1)
    assert (await calc_no_exp(3)) == 9
    assert calc_count == 1

    calc_count = 0

    @lru_cache(maxsize=2, expire_after=0.05)
    async def calc_exp(n):
        nonlocal calc_count
        await asyncio.sleep(0)
        calc_count += 1
        return n * n

    assert (await calc_exp(3)) == 9
    assert calc_count == 1
    assert (await calc_exp(3)) == 9
    assert calc_count == 1
    await asyncio.sleep(0.1)
    assert (await calc_exp(3)) == 9
    assert calc_count == 2



"""
A set of common utilities for legacy Pythons
"""

import asyncio
import functools

from .. import compat

_has_task_name = hasattr(asyncio.Task, 'get_name')


def create_task_with_name(coro, *, name=None):
    loop = compat.get_running_loop()
    if _has_task_name and name:
        child_task = loop.create_task(coro, name=name)
    else:
        child_task = loop.create_task(coro)
    return child_task


def patch_task(task) -> None:
    # In Python 3.8 we'll need proper API on asyncio.Task to
    # make TaskGroups possible. We need to be able to access
    # information about task cancellation, more specifically,
    # we need a flag to say if a task was cancelled or not.
    # We also need to be able to flip that flag.

    def _task_cancel(task, orig_cancel, msg=None):
        task.__cancel_requested__ = True
        if msg is not None:
            return orig_cancel(msg)
        else:
            return orig_cancel()

    if hasattr(task, '__cancel_requested__'):
        return

    task.__cancel_requested__ = False
    # confirm that we were successful at adding the new attribute:
    assert not task.__cancel_requested__

    orig_cancel = task.cancel
    task.cancel = functools.partial(_task_cancel, task, orig_cancel)



import logging
import os
from typing import Set

import aiotools
import zmq, zmq.asyncio

num_workers = 4

log_init_states: Set[str] = set()


def get_logger(name: str, pid: int) -> logging.Logger:
    log = logging.getLogger(name)
    # initialize only once for each logger identified by the name
    if name not in log_init_states:
        sh = logging.StreamHandler()
        fmt = logging.Formatter(
            f'%(relativeCreated).3f %(name)s[{pid}] %(levelname)s: %(message)s')
        sh.setFormatter(fmt)
        log.addHandler(sh)
        log.propagate = False
        log.setLevel(logging.INFO)
        log_init_states.add(name)
    return log


def router_main(_, pidx, args):
    log = get_logger('examples.zmqserver.extra', pidx)
    zctx = zmq.Context()
    zctx.linger = 0
    in_sock = zctx.socket(zmq.PULL)
    in_sock.bind('tcp://*:5033')
    out_sock = zctx.socket(zmq.PUSH)
    out_sock.bind('ipc://example-events')
    try:
        log.info('router proxy started')
        zmq.proxy(in_sock, out_sock)
    except KeyboardInterrupt:
        pass
    except Exception:
        log.exception('unexpected error')
    finally:
        for _ in range(num_workers):
            out_sock.send(b'')  # sentinel
        log.info('router proxy terminated')
        in_sock.close()
        out_sock.close()
        zctx.term()
        os.unlink('example-events')


@aiotools.actxmgr
async def worker_main(loop, pidx, args):
    log = get_logger('examples.zmqserver.worker', pidx)
    zctx = zmq.asyncio.Context()
    router = zctx.socket(zmq.PULL)
    router.connect('ipc://example-events')

    async def process_incoming(router):
        while True:
            data = await router.recv()
            if not data:
                return
            log.info(data)

    task = loop.create_task(process_incoming(router))
    log.info('started')

    try:
        yield
    finally:
        await task
        router.close()
        zctx.term()
        log.info('terminated')


if __name__ == '__main__':
    # This example must be run with multiprocessing.
    server = aiotools.start_server(
        worker_main,
        use_threading=False,
        num_workers=num_workers,
        extra_procs=[router_main],
        start_method='spawn',
    )



"""
Provides an implementation of asynchronous context manager and its applications.

.. note::

   The async context managers in this module are transparent aliases to
   ``contextlib.asynccontextmanager`` of the standard library in Python 3.7
   and later.
"""

import contextlib
import asyncio
from typing import Iterable, Optional, List

__all__ = [
    'AsyncContextManager', 'async_ctx_manager', 'actxmgr',
    'aclosing', 'closing_async',
    'AsyncContextGroup', 'actxgroup',
    'AsyncExitStack',
]


AbstractAsyncContextManager = contextlib.AbstractAsyncContextManager
AsyncContextManager = contextlib._AsyncGeneratorContextManager
AsyncExitStack = contextlib.AsyncExitStack
async_ctx_manager = contextlib.asynccontextmanager
aclosing = contextlib.aclosing


class closing_async:
    """
    An analogy to :func:`contextlib.closing` for objects with ``close()``
    methods as async functions.

    .. versionadded:: 1.5.6
    """

    def __init__(self, thing):
        self.thing = thing

    async def __aenter__(self):
        return self.thing

    async def __aexit__(self, *args):
        await self.thing.close()


class AsyncContextGroup:
    """
    Merges a group of context managers into a single context manager.
    Internally it uses :func:`asyncio.gather()` to execute them with overlapping,
    to reduce the execution time via asynchrony.

    Upon entering, you can get values produced by the entering steps from
    the passed context managers (those ``yield``-ed) using an ``as`` clause of
    the ``async with``
    statement.

    After exits, you can check if the context managers have finished
    successfully by ensuring that the return values of ``exit_states()`` method
    are ``None``.

    .. note::

       You cannot return values in context managers because they are
       generators.

    If an exception is raised before the ``yield`` statement of an async
    context manager, it is stored at the corresponding manager index in the
    as-clause variable.  Similarly, if an exception is raised after the
    ``yield`` statement of an async context manager, it is stored at the
    corresponding manager index in the ``exit_states()`` return value.

    Any exceptions in a specific context manager does not interrupt others;
    this semantic is same to ``asyncio.gather()``'s when
    ``return_exceptions=True``.  This means that, it is user's responsibility
    to check if the returned context values are exceptions or the intended ones
    inside the context body after entering.

    :param context_managers: An iterable of async context managers.
                             If this is ``None``, you may add async context
                             managers one by one using the :meth:`~.add`
                             method.

    """

    def __init__(self,
                 context_managers: Optional[Iterable[AbstractAsyncContextManager]] = None):  # noqa
        self._cm = list(context_managers) if context_managers else []
        self._cm_yields: List[asyncio.Task] = []
        self._cm_exits: List[asyncio.Task] = []

    def add(self, cm):
        """
        TODO: fill description
        """
        self._cm.append(cm)

    async def __aenter__(self):
        # Exceptions in context managers are stored into _cm_yields list.
        # NOTE: There is no way to "skip" the context body even if the entering
        #       process fails.
        self._cm_yields[:] = await asyncio.gather(
            *(e.__aenter__() for e in self._cm),
            return_exceptions=True)
        return self._cm_yields

    async def __aexit__(self, *exc_info):
        # Clear references to context variables.
        self._cm_yields.clear()
        # Exceptions are stored into _cm_exits list.
        self._cm_exits[:] = await asyncio.gather(
            *(e.__aexit__(*exc_info) for e in self._cm),
            return_exceptions=True)

    def exit_states(self):
        """
        TODO: fill description
        """
        return self._cm_exits


# Shorter aliases
actxmgr = async_ctx_manager
actxgroup = AsyncContextGroup



from types import TracebackType
from typing import Protocol


class AsyncExceptionHandler(Protocol):
    """
    A shorthand for an async exception handler type.
    This is always called under exception context where
    :func:`sys.exc_info()` is available.
    """
    async def __call__(
        self,
        exc_type: type[Exception],
        exc_obj: Exception,
        exc_tb: TracebackType,
    ) -> None:
        ...


class MultiError(ExceptionGroup):

    def __init__(self, msg: str, errors=()) -> None:
        super().__init__(msg, errors)
        self.__errors__ = errors

    def get_error_types(self) -> set[type[Exception]]:
        return {type(e) for e in self.exceptions}


class TaskGroupError(MultiError):
    pass



import pytest

import asyncio
from aiotools.defer import defer, adefer
from aiotools.func import apartial


def test_defer():

    x = []

    @defer
    def myfunc(defer):
        x.append(1)
        defer(lambda: x.append(1))
        x.append(2)
        defer(lambda: x.append(2))
        x.append(3)
        defer(lambda: x.append(3))

    myfunc()
    assert x == [1, 2, 3, 3, 2, 1]


def test_defer_inner_exception():

    x = []

    @defer
    def myfunc(defer):
        x.append(1)
        defer(lambda: x.append(1))
        x.append(2)
        defer(lambda: x.append(2))
        raise ZeroDivisionError
        x.append(3)
        defer(lambda: x.append(3))

    with pytest.raises(ZeroDivisionError):
        myfunc()
    assert x == [1, 2, 2, 1]


def test_defer_wrong_func():
    with pytest.raises(AssertionError):
        @defer
        async def myfunc(defer):
            pass


@pytest.mark.asyncio
async def test_adefer():

    x = []

    @adefer
    async def myfunc(defer):
        x.append(1)
        defer(lambda: x.append(1))
        x.append(2)
        defer(lambda: x.append(2))
        x.append(3)
        defer(lambda: x.append(3))

    await myfunc()
    assert x == [1, 2, 3, 3, 2, 1]


def test_adefer_wrong_func():
    with pytest.raises(AssertionError):
        @adefer
        def myfunc(defer):
            pass


@pytest.mark.asyncio
async def test_adefer_coro():

    x = []

    async def async_append(target, item):
        target.append(item)
        await asyncio.sleep(0)

    @adefer
    async def myfunc(defer):
        x.append(1)
        defer(async_append(x, 1))
        x.append(2)
        defer(async_append(x, 2))
        x.append(3)
        defer(async_append(x, 3))

    await myfunc()
    assert x == [1, 2, 3, 3, 2, 1]


@pytest.mark.asyncio
async def test_adefer_corofunc():

    x = []

    async def async_append(target, item):
        target.append(item)
        await asyncio.sleep(0)

    @adefer
    async def myfunc(defer):
        x.append(1)
        defer(apartial(async_append, x, 1))
        x.append(2)
        defer(apartial(async_append, x, 2))
        x.append(3)
        defer(apartial(async_append, x, 3))

    await myfunc()
    assert x == [1, 2, 3, 3, 2, 1]


@pytest.mark.asyncio
async def test_adefer_inner_exception():

    x = []

    async def async_append(target, item):
        target.append(item)
        await asyncio.sleep(0)

    @adefer
    async def myfunc(defer):
        x.append(1)
        defer(apartial(async_append, x, 1))
        x.append(2)
        defer(apartial(async_append, x, 2))
        raise ZeroDivisionError
        x.append(3)
        defer(apartial(async_append, x, 3))

    with pytest.raises(ZeroDivisionError):
        await myfunc()
    assert x == [1, 2, 2, 1]



from .base import TaskGroup, TaskGroupError, current_taskgroup
from .persistent import PersistentTaskGroup, current_ptaskgroup
from .types import MultiError

__all__ = (
    "MultiError",
    "TaskGroup",
    "TaskGroupError",
    "current_taskgroup",
    "PersistentTaskGroup",
    "current_ptaskgroup",
)



import asyncio
import sys
import warnings

import pytest

from aiotools import (
    TaskGroup,
    TaskGroupError,
    VirtualClock,
)


@pytest.mark.asyncio
async def test_taskgroup_naming():

    async def subtask():
        pass

    async with TaskGroup(name="XYZ") as tg:
        t = tg.create_task(subtask(), name="ABC")
        assert tg.get_name() == "XYZ"
        if hasattr(t, 'get_name'):
            assert t.get_name() == "ABC"


@pytest.mark.asyncio
async def test_delayed_subtasks():
    with VirtualClock().patch_loop():
        async with TaskGroup() as tg:
            t1 = tg.create_task(asyncio.sleep(3, 'a'))
            t2 = tg.create_task(asyncio.sleep(2, 'b'))
            t3 = tg.create_task(asyncio.sleep(1, 'c'))
        assert t1.done()
        assert t2.done()
        assert t3.done()
        assert t1.result() == 'a'
        assert t2.result() == 'b'
        assert t3.result() == 'c'


@pytest.mark.asyncio
@pytest.mark.skipif(
    sys.version_info < (3, 7),
    reason='contextvars is available only in Python 3.7 or later',
)
async def test_contextual_taskgroup():

    from aiotools import current_taskgroup

    refs = []

    async def check_tg(delay):
        await asyncio.sleep(delay)
        refs.append(current_taskgroup.get())

    with VirtualClock().patch_loop():
        async with TaskGroup() as outer_tg:
            ot1 = outer_tg.create_task(check_tg(0.1))
            async with TaskGroup() as inner_tg:
                it1 = inner_tg.create_task(check_tg(0.2))
            ot2 = outer_tg.create_task(check_tg(0.3))
        assert ot1.done()
        assert ot2.done()
        assert it1.done()
        assert refs == [outer_tg, inner_tg, outer_tg]

        with pytest.raises(LookupError):
            # outside of any taskgroup, this is an error.
            current_taskgroup.get()


@pytest.mark.skipif(
    sys.version_info < (3, 7),
    reason='contextvars is available only in Python 3.7 or later',
)
@pytest.mark.filterwarnings('ignore::RuntimeWarning')
@pytest.mark.asyncio
async def test_contextual_taskgroup_spawning():

    from aiotools import current_taskgroup

    total_jobs = 0

    async def job():
        nonlocal total_jobs
        await asyncio.sleep(0)
        total_jobs += 1

    async def spawn_job():
        await asyncio.sleep(0)
        tg = current_taskgroup.get()
        tg.create_task(job())

    async def inner_tg_job():
        await asyncio.sleep(0)
        async with TaskGroup() as tg:
            tg.create_task(job())

    with VirtualClock().patch_loop():

        total_jobs = 0
        async with TaskGroup() as tg:
            t = tg.create_task(spawn_job())
        assert t.done()
        del t
        assert total_jobs == 1

        total_jobs = 0
        async with TaskGroup() as tg:
            tg.create_task(inner_tg_job())
            tg.create_task(spawn_job())
            tg.create_task(inner_tg_job())
            tg.create_task(spawn_job())
            # Give the subtasks chances to run.
            await asyncio.sleep(1)
        assert total_jobs == 4


@pytest.mark.asyncio
async def test_taskgroup_cancellation():
    with VirtualClock().patch_loop():

        async def do_job(delay, result):
            # NOTE: replacing do_job directly with asyncio.sleep
            #       results future-pending-after-loop-closed error,
            #       because asyncio.sleep() is not a task but a future.
            await asyncio.sleep(delay)
            return result

        with pytest.raises(asyncio.CancelledError):
            async with TaskGroup() as tg:
                t1 = tg.create_task(do_job(0.3, 'a'))
                t2 = tg.create_task(do_job(0.6, 'b'))
                await asyncio.sleep(0.5)
                raise asyncio.CancelledError

        assert t1.done()
        assert t2.cancelled()
        assert t1.result() == 'a'


@pytest.mark.asyncio
async def test_subtask_cancellation():

    results = []

    async def do_job():
        await asyncio.sleep(1)
        results.append('a')

    async def do_cancel():
        await asyncio.sleep(0.5)
        raise asyncio.CancelledError

    with VirtualClock().patch_loop():
        async with TaskGroup() as tg:
            t1 = tg.create_task(do_job())
            t2 = tg.create_task(do_cancel())
            t3 = tg.create_task(do_job())
        assert t1.done()
        assert t2.cancelled()
        assert t3.done()
        assert results == ['a', 'a']


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "cancel_msg",
    (
        ["MISSING", None, "msg"]
        if sys.version_info >= (3, 9, 0)
        else ["MISSING"]
    ),
)
async def test_cancel_parent_task(cancel_msg):

    results = []

    async def do_job():
        await asyncio.sleep(1)
        results.append('a')

    async def parent():
        async with TaskGroup() as tg:
            tg.create_task(do_job())

    with VirtualClock().patch_loop():
        parent_task = asyncio.ensure_future(parent())
        await asyncio.sleep(0.1)
        if cancel_msg == "MISSING":
            parent_task.cancel()
        else:
            parent_task.cancel(cancel_msg)
        await asyncio.gather(parent_task, return_exceptions=True)
        assert results == []


@pytest.mark.asyncio
async def test_taskgroup_distinguish_inner_error_and_outer_cancel():

    async def do_error():
        await asyncio.sleep(0.5)
        raise ValueError("bad stuff")

    with VirtualClock().patch_loop():

        with pytest.raises(TaskGroupError) as eg:
            async with TaskGroup() as tg:
                t1 = tg.create_task(do_error())
                # The following sleep is cancelled due to do_error(),
                # raising CancelledError!
                await asyncio.sleep(1)
                # We need to preserve the source exception instead of the outer
                # cancellation.  __aexit__() should not treat CancelledError as a
                # base error to handle such cases.

        assert t1.done()
        assert not t1.cancelled()
        assert isinstance(t1.exception(), ValueError)
        assert len(eg.value.__errors__) == 1
        assert isinstance(eg.value.__errors__[0], ValueError)


@pytest.mark.asyncio
async def test_taskgroup_error():
    with VirtualClock().patch_loop():

        async def do_job(delay, result):
            await asyncio.sleep(delay)
            if result == 'x':
                raise ZeroDivisionError('oops')
            else:
                return 99

        with pytest.raises(TaskGroupError) as e:
            async with TaskGroup() as tg:
                t1 = tg.create_task(do_job(0.3, 'a'))
                t2 = tg.create_task(do_job(0.5, 'x'))
                t3 = tg.create_task(do_job(0.7, 'a'))

        assert len(e.value.__errors__) == 1
        assert type(e.value.__errors__[0]).__name__ == 'ZeroDivisionError'

        assert t1.done()
        assert await t1 == 99
        assert t1.result() == 99
        assert t1.exception() is None

        assert t2.done()
        with pytest.raises(ZeroDivisionError):
            await t2
        with pytest.raises(ZeroDivisionError):
            t2.result()
        assert type(t2.exception()).__name__ == 'ZeroDivisionError'

        assert t3.cancelled()


@pytest.mark.asyncio
async def test_taskgroup_error_weakref():
    with VirtualClock().patch_loop():

        results = []

        async def do_job(delay, result):
            await asyncio.sleep(delay)
            if result == 'x':
                results.append('x')
                raise ZeroDivisionError('oops')
            else:
                results.append('o')
                return 99

        with pytest.raises(TaskGroupError) as e:
            async with TaskGroup() as tg:
                # We don't keep the reference to the tasks,
                # but they should behave the same way
                # regardless of usage of WeakSet in the implementation.
                tg.create_task(do_job(0.3, 'a'))
                tg.create_task(do_job(0.5, 'x'))
                tg.create_task(do_job(0.7, 'a'))

        assert len(e.value.__errors__) == 1
        assert type(e.value.__errors__[0]).__name__ == 'ZeroDivisionError'
        assert results == ['o', 'x']


@pytest.mark.asyncio
async def test_taskgroup_memoryleak_with_persistent_tg():

    with VirtualClock().patch_loop(), \
         warnings.catch_warnings():
        warnings.simplefilter("ignore")

        async def do_job(delay):
            await asyncio.sleep(delay)
            return 1

        async with TaskGroup() as tg:
            for count in range(1000):
                await asyncio.sleep(1)
                tg.create_task(do_job(10))
                if count == 100:
                    # 10 ongoing tasks + 1 just spawned task
                    assert len(tg._tasks) == 11
            await asyncio.sleep(10.1)
            assert len(tg._tasks) == 0



import asyncio
import aiotools

lock = asyncio.Lock()


@aiotools.actxmgr
async def mygen(input_value):
    print(input_value)
    await lock.acquire()
    print('The lock is acquired.')
    try:
        yield 'return_value'
    finally:
        lock.release()
        print('The lock is released.')


async def run():
    try:
        async with mygen('input_value') as return_value:
            print(return_value)
            raise RuntimeError
    except RuntimeError:
        print('RuntimeError is caught!')  # you can catch exceptions here.


if __name__ == '__main__':
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    try:
        loop.run_until_complete(run())
    finally:
        loop.stop()



import asyncio

import pytest

import aiotools


@pytest.mark.asyncio
async def test_timer():
    """
    Test the timer functionality.
    """
    vclock = aiotools.VirtualClock()
    with vclock.patch_loop():
        count = 0

        async def counter(interval):
            assert interval == 0.1
            nonlocal count
            await asyncio.sleep(0)
            count += 1

        count = 0
        timer = aiotools.create_timer(counter, 0.1)
        await asyncio.sleep(0.22)
        timer.cancel()
        await timer
        assert count == 3

        count = 0
        timer = aiotools.create_timer(counter, 0.1, aiotools.TimerDelayPolicy.CANCEL)
        await asyncio.sleep(0.22)
        timer.cancel()
        await timer
        # should have same results
        assert count == 3


@pytest.mark.asyncio
async def test_timer_leak_default():
    """
    Test if the timer-fired tasks are claned up properly
    even when each timer-fired task takes longer than the timer interval.
    (In this case they will accumulate indefinitely!)
    """
    vclock = aiotools.VirtualClock()
    with vclock.patch_loop():
        spawn_count = 0
        cancel_count = 0
        done_count = 0

        async def delayed(interval):
            nonlocal spawn_count, cancel_count, done_count
            spawn_count += 1
            try:
                await asyncio.sleep(5)
                done_count += 1
            except asyncio.CancelledError:
                cancel_count += 1

        task_count = len(aiotools.compat.all_tasks())
        timer = aiotools.create_timer(delayed, 1)
        await asyncio.sleep(9.9)
        timer.cancel()
        await timer
        assert task_count + 1 >= len(aiotools.compat.all_tasks())
        assert spawn_count == done_count + cancel_count
        assert spawn_count == 10
        assert cancel_count == 5


@pytest.mark.asyncio
async def test_timer_leak_cancel():
    """
    Test the effect of TimerDelayPolicy.CANCEL which always
    cancels any pending previous tasks on each interval.
    """
    vclock = aiotools.VirtualClock()
    with vclock.patch_loop():
        spawn_count = 0
        cancel_count = 0
        done_count = 0

        async def delayed(interval):
            nonlocal spawn_count, cancel_count, done_count
            spawn_count += 1
            try:
                await asyncio.sleep(1)
            except asyncio.CancelledError:
                cancel_count += 1
            else:
                done_count += 1

        task_count = len(aiotools.compat.all_tasks())
        timer = aiotools.create_timer(
            delayed, 0.01, aiotools.TimerDelayPolicy.CANCEL,
        )
        await asyncio.sleep(0.1)
        timer.cancel()
        await timer
        await asyncio.sleep(0)
        assert task_count + 1 >= len(aiotools.compat.all_tasks())
        assert spawn_count == cancel_count + done_count
        assert cancel_count == 10
        assert done_count == 0


@pytest.mark.asyncio
async def test_timer_leak_nocancel():
    """
    Test the effect of TimerDelayPolicy.CANCEL which always
    cancels any pending previous tasks on each interval.
    """
    vclock = aiotools.VirtualClock()
    with vclock.patch_loop():
        spawn_count = 0
        cancel_count = 0
        done_count = 0

        async def delayed(interval):
            nonlocal spawn_count, cancel_count, done_count
            spawn_count += 1
            try:
                await asyncio.sleep(0)
            except asyncio.CancelledError:
                cancel_count += 1
            else:
                done_count += 1

        task_count = len(aiotools.compat.all_tasks())
        timer = aiotools.create_timer(
            delayed, 0.01, aiotools.TimerDelayPolicy.CANCEL,
        )
        await asyncio.sleep(0.096)
        timer.cancel()
        await timer
        await asyncio.sleep(0)
        assert task_count + 1 >= len(aiotools.compat.all_tasks())
        assert spawn_count == cancel_count + done_count
        assert cancel_count == 0
        assert done_count == 10
